/*=======================================================
  Author:   R. Muth
  Version:  $Id: eval.c,v 1.30 1998/12/09 23:08:38 muth Exp muth $
  Comment:  This file contains the constant propagation and folding code
=======================================================*/


#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"

/*=======================================================
=======================================================*/

/*
  these are the initial suggestions for the lattice top and bottom
  not that both are odd, so they cannot be a relocatable address!
  */
   
static INT64 LatticeTop = 0x3141592653589793;
static INT64 LatticeBot = ~0x3141592653589793;


#define ABBL_EVAL_IN_VALUE(i)       ABBL_U1(i).quadword
#define ABBL_EVAL_DEF_VALUE(i)      ABBL_U2(i).quadword
#define ABBL_EVAL_DEF_VALUE_TRUE(i) ABBL_U3(i).quadword


/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "RegisterForward"

extern NUMBER RegisterForward(INDEX iins)
{
    static NUMBER count = 0;

        
    NUMBER change = 0;

    if( (AINS_FLAGS(iins) & FI_EVAL_A) &&
        AINS_REG_A(iins) != REG_ZERO &&
        AINS_REG_A(iins) != REG_FZERO )
    {
        if( AINS_A_EVAL(iins) == 0 )
        {
            if( AINS_REG_A(iins) >= 32 )
                AINS_REG_A(iins) = REG_FZERO;
            else
                AINS_REG_A(iins) = REG_ZERO;
            AINS_FLAGS(iins) |= FI_CHANGED;
            change = 1;
            count++;
        }
    }

    if( (AINS_FLAGS(iins) & FI_EVAL_B) &&
        AINS_REG_B(iins) != REG_ZERO &&
        AINS_REG_B(iins) != REG_FZERO )
    {
        if( AINS_B_EVAL(iins) == 0 )
        {
            if( AINS_REG_B(iins) == REG_LIT )
                AINS_REG_B(iins) = REG_ZERO;
            else if( AINS_REG_B(iins) >= 32 )
                AINS_REG_B(iins) = REG_FZERO;
            else
                AINS_REG_B(iins) = REG_ZERO;
            AINS_FLAGS(iins) |= FI_CHANGED;
            count++;
            change = 1;            
        }
        else if( AINS_REG_B(iins) != REG_LIT && AINS_B_EVAL(iins) > 0 && AINS_B_EVAL(iins) <= 255 ) 
        {
            if( AINS_TYPE(iins) == IT_ICM ||
                AINS_TYPE(iins) == IT_IOP ||
                AINS_TYPE(iins) == IT_ICP ||
                AINS_TYPE(iins) == IT_MLL ||
                AINS_TYPE(iins) == IT_MLQ ||
                AINS_TYPE(iins) == IT_MLH ||
                AINS_TYPE(iins) == IT_EXT ||
                AINS_TYPE(iins) == IT_SFT ||
                AINS_TYPE(iins) == IT_LOG )
            {
                AINS_DATA(iins) = AINS_B_EVAL(iins);
                AINS_REG_B(iins) = REG_LIT;
                AINS_FLAGS(iins) |= FI_CHANGED;
                count++;
                change = 1;
            }
        }
        
    }

    /* this comparison to the integer zero is on purpose */
    /* because there are two different float zeros */
            
    return change;
}
/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "StrRedUnaligned"

static NUMBER StrRedUnaligned(INDEX iins)
{
    if( AINS_TYPE(iins) != I_ldq_u && AINS_TYPE(iins) != I_stq_u )
        return 0;

    if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        HURRAH(("Unaligned ld/st to/from %lx NOT IMPLEMENTED YET (%ld\n)\n",AINS_B_EVAL(iins),ABBL_COUNT(AINS_BBL(iins))));
        PrintIns(stdout,iins);
    }
    return 0;
}
/*=======================================================
=======================================================*/

    
#undef __PROC__
#define __PROC__ "StrRedICM"

static NUMBER StrRedICM(INDEX iins)
{
    if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        BOOL taken;

        switch(AINS_CODE(iins))
        {
        case I_cmoveq:   taken = AINS_A_EVAL(iins) ? 0 : 1; break;
        
        case I_cmovne:   taken = AINS_A_EVAL(iins) ? 1 : 0; break;

        case I_cmovlt:   taken = AINS_A_EVAL(iins) < 0 ? 1 : 0; break;
        
        case I_cmovge:   taken = AINS_A_EVAL(iins) < 0 ? 0 : 1; break;
        
        case I_cmovgt:   taken = AINS_A_EVAL(iins) > 0 ? 1 : 0; break;

        case I_cmovle:   taken = AINS_A_EVAL(iins) > 0 ? 0 : 1; break;
            
        case I_cmovlbs:  taken = AINS_A_EVAL(iins) & 1 ? 1 : 0; break;
            
        case I_cmovlbc:  taken = AINS_A_EVAL(iins) & 1 ? 0 : 1; break;

        default:
            taken = 1;
            ERROR(("Unexpected ICM instruction\n"));
            PrintIns(stdout,iins);
            break;
        }

        if( taken )
        {
#if 0
        printf("Transforming ICM\n");
        PrintIns(stdout,iins);
#endif
        if( AINS_REG_B(iins) == REG_LIT )
            AinsMakeLda(iins,REG_ZERO,AINS_REG_C(iins),AINS_DATA(iins));
        else
            AinsMakeLda(iins,AINS_REG_B(iins),AINS_REG_C(iins),0);
#if 0
        PrintIns(stdout,iins);
#endif
            return 1;
        }
        else
        {
#if 0
        printf("Killing ICM\n");
        PrintIns(stdout,iins);
#endif
            AinsKill(iins);
            return 1;
        }

    }

    if( AINS_REG_B(iins) == AINS_REG_C(iins) )
    {
#if 0
        printf("Killing ICM\n");
        PrintIns(stdout,iins);
#endif
        AinsKill(iins);
        return 1;
    }

    return 0;
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "StrRedFCM"

static NUMBER StrRedFCM(INDEX iins)
{
    if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        BOOL taken;

        switch(AINS_CODE(iins))
        {
        case I_fcmoveq:   taken = AINS_A_EVAL(iins) == 0.0 ? 1 : 0; break;
        
        case I_fcmovne:   taken = AINS_A_EVAL(iins) == 0.0 ? 0 : 1; break;

        case I_fcmovlt:   taken = AINS_A_EVAL(iins) <  0.0 ? 1 : 0; break;
        
        case I_fcmovge:   taken = AINS_A_EVAL(iins) <  0.0 ? 0 : 1; break;
        
        case I_fcmovgt:   taken = AINS_A_EVAL(iins) >  0.0 ? 1 : 0; break;

        case I_fcmovle:   taken = AINS_A_EVAL(iins) >  0.0 ? 0 : 1; break;
            
        default:
            taken = 1;
            ERROR(("Unexpected FCM instruction\n"));
            PrintIns(stdout,iins);
            break;
        }

        if( taken )
        {
            AinsMakeMove(iins,AINS_REG_B(iins),DT_TFLT,AINS_REG_C(iins),DT_TFLT);
            return 1;
        }
        else
        {
            AinsKill(iins);
            return 1;
        }

    }

    if( AINS_REG_B(iins) == AINS_REG_C(iins) )
    {
        AinsKill(iins);
        return 1;
    }

    
    return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "KillBadLoad"

static void KillBadLoad(INDEX iins)
{
    INDEX ibbl = AINS_BBL(iins);
    /*
    INDEX ifun = ABBL_FUN(ibbl);
    */

    if( AbblNpreds(ibbl) == 0 )
    {
        AbblKill(ibbl);
        return;
    }

    /* NEEDS WORK: this is a bad hack bad works very well */

    if( HasOption(PRINT_OPT_VERBOSE,"Eval") )
    {
        WARNING(("Killing LOAD from bad adr %016lx\n",AINS_B_EVAL(iins)));
        PrintIns(stdout,iins);
        PrintBblSuper(stdout,ibbl-1);
        PrintBblSuper(stdout,ibbl);
        PrintBblSuper(stdout,ibbl+1);
    }
    
    AinsKill(iins);
}

    

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "KillBadJump"

static void KillBadJump(INDEX iins)
{
    INDEX jbbl;
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);

    if( AbblNpreds(ibbl) == 0 )
    {
        AbblKill(ibbl);
        return;
    }
    
    /* NEEDS WORK: this is a bad hack bad works very well */

    if( HasOption(PRINT_OPT_VERBOSE,"Eval") )
    {
        WARNING(("Killing JMP to bad adr %016lx\n",AINS_B_EVAL(iins)));
        PrintIns(stdout,iins);
        PrintBblSuper(stdout,ibbl);
    }
    
    if( AINS_CODE(iins) == I_jsr )
    {
        CASSERT( ABBL_TYPE(ibbl)==BT_CALL,("jsr found in non bt_call bbl\n"));
        CASSERT( VALID(ABBL_EDG_CALL(ibbl)),("bt_call bbl with invalid edg_call\n"));
        CASSERT( VALID(ABBL_EDG_LINK(ibbl)),("bt_call bbl with invalid edg_link\n"));

        jbbl = AEDG_TAIL( ABBL_EDG_LINK(ibbl) );
        AinsKill(iins);
        AedgKillFancy(ABBL_EDG_CALL(ibbl));
        AedgKillFancy(ABBL_EDG_LINK(ibbl));
        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_NORM));                                   
    }
    else if( AINS_CODE(iins) == I_jmp )
    {
        CASSERT( ABBL_TYPE(ibbl) == BT_JUMP,("jmp found in non bt_jump bbl\n"));
        CASSERT( VALID(ABBL_EDG_JUMP(ibbl)),("bt_jump bbl with invalid edg_jump\n"));
        
        AinsKill(iins);
        AedgKillFancy(ABBL_EDG_JUMP(ibbl));        
        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_LAST(ifun),ET_NORM));                                   
    }
    else
    {
        ERROR(("Unexpected Jump Instruction (target: 0x%lx)\n",AINS_B_EVAL(iins)));
        PrintBblSuper(stdout,ibbl);
#if 0
        PrintFunSuper(stdout,ABBL_FUN(ibbl));
#endif
        
    }
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "StrRedJMP"

static NUMBER StrRedJMP(INDEX iins)
{
    INDEX ibbl;
    INDEX isec;
    INDEX target;

    if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        return 0;

    ibbl = AINS_BBL(iins);
    isec = LocateSection(AINS_B_EVAL(iins));
    if( isec != SysGetTextSection() )
    {
        KillBadJump(iins);
        return 0;
    }
    
    /* target iins might be killed because it is a useless
       instrucion but its bbl should be live */

    
    switch( AINS_CODE(iins) )
    {
    case I_jsr:
        CASSERT( ABBL_TYPE(ibbl)==BT_CALL,("jsr found in non bt_call bbl\n"));
        CASSERT( VALID(ABBL_EDG_CALL(ibbl)),("bt_call bbl with invalid edg_call\n"));        

        target = FindTargetBbl(AINS_B_EVAL(iins));
        AinsMakeBsr(iins,AINS_REG_C(iins));
        AINS_C_EVAL(iins) = Ins2Addr(iins+1);    /* NEEDS WORK */
        AINS_FLAGS(iins) |= FI_EVAL_C;
        AedgKillFancy(ABBL_EDG_CALL(ibbl));
        AedgLinkFancy( AedgNewFancy(ibbl,target,ET_FUN_CALL));
#if 0
        printf("resolved jsr from %d [%s] to %d [%s]\n",
               ABBL_FUN(ibbl),AFUN_NAME(ABBL_FUN(ibbl)),ABBL_FUN(target),AFUN_NAME(ABBL_FUN(target)));
#endif        
        return 1;

    case I_jmp:
        CASSERT( ABBL_TYPE(ibbl)==BT_JUMP,("jmp found in non bt_jump bbl\n"));
        CASSERT( VALID(ABBL_EDG_JUMP(ibbl)),("bt_jump bbl with invalid edg_jump\n"));

        target = FindTargetBbl(AINS_B_EVAL(iins));
        AinsKill(iins);
        AedgKillFancy(ABBL_EDG_JUMP(ibbl));
        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy( AedgNewFancy(ibbl,target,ET_NORM));
        return 1;
        
    case I_ret:
        return 0;
            
    default:
        ERROR(("Did not evaluate jsr_co\n"));
        return 0;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "StrRedIBR"


static NUMBER StrRedIBR(INDEX iins)  
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX iedg;
    BOOL taken;

    CASSERT( ABBL_TYPE(ibbl) == BT_BRANCH,("ibr found in non bt_branch bbl\n"));
    CASSERT( VALID(ABBL_EDG_TRUE(ibbl)),("bt_branch bbl with invalid edg_true\n"));
    CASSERT( VALID(ABBL_EDG_FALSE(ibbl)),("bt_branch bbl with invalid edg_false\n"));

    if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
        return 0;    

    switch(AINS_CODE(iins))
    {
    case I_beq:   taken =  AINS_A_EVAL(iins) == 0 ? 1 : 0; break;
        
    case I_bne:   taken =  AINS_A_EVAL(iins) == 0 ? 0 : 1; break;
        
    case I_blt:   taken =  AINS_A_EVAL(iins) <  0 ? 1 : 0; break;
        
    case I_bge:   taken =  AINS_A_EVAL(iins) <  0 ? 0 : 1; break;
        
    case I_bgt:   taken =  AINS_A_EVAL(iins) >  0 ? 1 : 0; break;
        
    case I_ble:   taken =  AINS_A_EVAL(iins) >  0 ? 0 : 1; break;

    case I_blbs:  taken =  AINS_A_EVAL(iins) &  1 ? 1 : 0; break;
        
    case I_blbc:  taken =  AINS_A_EVAL(iins) &  1 ? 0 : 1; break;
        
    default:      taken = 1; FATAL(("weird\n"));
    }
    
    if( taken )
    {
        iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_TRUE(ibbl)),ET_NORM);
    }
    else
    {
        iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_FALSE(ibbl)),ET_NORM);
    }

    AedgKillFancy(ABBL_EDG_TRUE(ibbl));
    AedgKillFancy(ABBL_EDG_FALSE(ibbl));
    AinsKill(iins);
    ABBL_TYPE(ibbl) = BT_NORM;
    AedgLinkFancy(iedg);
    return 1;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "StrRedFBR"


static NUMBER StrRedFBR(INDEX iins)  
{
    INDEX ibbl = AINS_BBL(iins);
    INDEX iedg;
#if 0
    INDEX ifun;
#endif
    BOOL taken;
    
    CASSERT( ABBL_TYPE(ibbl) == BT_BRANCH,("fbr found in non bt_branch bbl\n"));
    CASSERT( VALID(ABBL_EDG_TRUE(ibbl)),("bt_branch bbl with invalid edg_true\n"));
    CASSERT( VALID(ABBL_EDG_FALSE(ibbl)),("bt_branch bbl with invalid edg_false\n"));


    if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
        return 0;    

    switch(AINS_CODE(iins))
    {
    case I_fbeq:   taken = AINS_A_FEVAL(iins) == 0.0 ? 1 : 0; break;
        
    case I_fbne:   taken = AINS_A_FEVAL(iins) == 0.0 ? 0 : 1; break;

    case I_fblt:   taken = AINS_A_FEVAL(iins) <  0.0 ? 1 : 0; break;
        
    case I_fbge:   taken = AINS_A_FEVAL(iins) <  0.0 ? 0 : 1; break;
        
    case I_fbgt:   taken = AINS_A_FEVAL(iins) >  0.0 ? 1 : 0; break;

    case I_fble:   taken = AINS_A_FEVAL(iins) >  0.0 ? 0 : 1; break;

    default:      taken = 1; FATAL(("weird\n"));
    }

#if 0
    ifun = ABBL_FUN(AINS_BBL(iins));
    HURRAH(("Evaluated cond branch  in fun %d [%s] - %ld\n",ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
    PrintIns(stdout,iins);
#endif
    
    if( taken )
    {
        iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_TRUE(ibbl)),ET_NORM);
    }
    else
    {
        iedg = AedgNewFancy(ibbl,AEDG_TAIL(ABBL_EDG_FALSE(ibbl)),ET_NORM);
    }

    AedgKillFancy(ABBL_EDG_TRUE(ibbl));
    AedgKillFancy(ABBL_EDG_FALSE(ibbl));
    AinsKill(iins);
    ABBL_TYPE(ibbl) = BT_NORM;
    AedgLinkFancy(iedg);
    return 1;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "StrRedMUL"


static NUMBER StrRedMUL(INDEX iins)  
{
#if 0
    INDEX ibbl = AINS_BBL(iins);
    INDEX ifun = ABBL_FUN(ibbl);

    if( !(AINS_FLAGS(iins) & (FI_EVAL_A|FI_EVAL_B)) )
        return 0;    

    if( AINS_FLAGS(iins) & FI_EVAL_A )
    {
        if( BinaryBits(AINS_A_EVAL(iins)) <= 2 || BinaryBits(-AINS_A_EVAL(iins)) <= 2 ||
            (AINS_A_EVAL(iins) < 100 && AINS_A_EVAL(iins) > -100) )
        {
            HURRAH(("partially constant mult [%s] %d in %d [%s] - %ld\n",
                    AINS_NAME(iins),AINS_A_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
                PrintIns(stdout,iins);
        }
    }
    
    if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        if( BinaryBits(AINS_B_EVAL(iins)) <= 2 || BinaryBits(-AINS_B_EVAL(iins)) <= 2 ||
            (AINS_B_EVAL(iins) < 100 && AINS_B_EVAL(iins) > -100) )
            {
                HURRAH(("partially constant mult [%s] %d in %d [%s] - %ld\n",
                        AINS_NAME(iins),AINS_B_EVAL(iins),ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
                PrintIns(stdout,iins);
            }
        
    }
#endif    
    return 0;
}

/*=======================================================
=======================================================*/
/*=======================================================
  =======================================================*/
#undef __PROC__
#define __PROC__ "EvalLDA"

static NUMBER EvalLDA(INDEX ii)
{
    if( AINS_FLAGS(ii) & FI_EVAL_B )
    {
        if( AINS_CODE(ii) == I_lda )
            AINS_C_EVAL(ii) = AINS_B_EVAL(ii) + AINS_DATA(ii);
        
        if( AINS_CODE(ii) == I_ldah )
            AINS_C_EVAL(ii) = AINS_B_EVAL(ii) + ((INT64)AINS_DATA(ii)<<16L);

        AINS_FLAGS(ii) |= FI_EVAL_C;
        return 1;
    }
    else
    {
        return 0;
    }
    
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "EvalILD"

static NUMBER EvalILD(INDEX iins)
{
    if( AINS_FLAGS(iins) & FI_EVAL_B )
    {
        INT64 address = AINS_B_EVAL(iins) + AINS_DATA(iins);
        INDEX isec = LocateSection(address);

        if( isec < 0 )
        {
            KillBadLoad(iins);
            return 1;
        }

        if( AsecIsReadOnly(isec) )
        {
            if( AsecIsCode(isec) )
            {
                ERROR(("read from code segment\n"));
                PrintBblSuper(stdout,AINS_BBL(iins));
                return 0;
            }
            

            if( AINS_CODE(iins) == I_ldq_u )
            {
                INDEX ibbl = AINS_BBL(iins);
                INDEX ifun = ABBL_FUN(ibbl);
                HURRAH(("ldq_u resolved %d [%s] - %ld\n",
                        ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            }
            AINS_C_EVAL(iins) = EmulILD(AINS_CODE(iins),AsecMapAddress(isec,address));
            AINS_FLAGS(iins) |= FI_EVAL_C;
            return 1;
        }

        
    }
    
    return 0;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "EvalFLD"

static NUMBER EvalFLD(INDEX ii)
{
    if( AINS_FLAGS(ii) & FI_EVAL_B )
    {
        INT64 address = AINS_B_EVAL(ii) + AINS_DATA(ii);
        INDEX is = LocateSection(address);

        if( is < 0 )
        {
            KillBadLoad(ii);
            return 1;
        }

        if( AsecIsReadOnly(is) )
        {
            CERROR( AsecIsCode(is),("read from code segment\n"));
            AINS_C_FEVAL(ii) = EmulFLD(AINS_CODE(ii),AsecMapAddress(is,address));
            AINS_FLAGS(ii) |= FI_EVAL_C;
            return 1;
        }
    }
    
    return 0;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "EvalIOP"

static NUMBER EvalIOP(INDEX ii)
{
    if( (AINS_FLAGS(ii) & FI_EVAL_B) && (AINS_FLAGS(ii) & FI_EVAL_A) )
    {
        AINS_C_EVAL(ii) = EmulIOP(AINS_CODE(ii),AINS_A_EVAL(ii),AINS_B_EVAL(ii));
        AINS_FLAGS(ii) |= FI_EVAL_C;
        return 1;
    }
    else
    {
        return 0;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "EvalFOP"

static NUMBER EvalFOP(INDEX ii)
{
    if( (AINS_FLAGS(ii) & FI_EVAL_B) && (AINS_FLAGS(ii) & FI_EVAL_A) )
    {
        AINS_C_FEVAL(ii) = EmulFOP(AINS_CODE(ii),AINS_A_FEVAL(ii),AINS_B_FEVAL(ii));
        AINS_FLAGS(ii) |= FI_EVAL_C;
#if 0
        if( AINS_CODE(ii) != I_cpys )
        {
            INDEX ibbl = AINS_BBL(ii);
            INDEX ifun = ABBL_FUN(ibbl);
            HURRAH(("Evaluated: %s %g %g = %g  in fun %d [%s] - %ld\n",
                    AINS_NAME(ii),AINS_A_FEVAL(ii),AINS_B_FEVAL(ii),AINS_C_FEVAL(ii),
                    ifun,AFUN_NAME(ifun),ABBL_COUNT(ibbl)));
            PrintIns(stdout,ii);
        }
        
#endif
        return 1;
    }
    else
    {
        return 0;
    }
}

/*=======================================================
FIX POINT COMPUTATION CODE
=======================================================*/


/*=======================================================
A new value has reached a bbl.
update the new in value with the join of the old and new value
if the in value changes, mark the bbl.
=======================================================*/

#undef __PROC__
#define __PROC__ "EvalLatticeJoin"

static void EvalLatticJoin(INDEX ibbl, INT64 value, INT64 LatticeBot, INT64 LatticeTop)
{
    INT64 old_value = ABBL_EVAL_IN_VALUE(ibbl);
    INT64 new_value;


    if( value == LatticeBot )
    {
        new_value = old_value;
    }
    else if( value  == LatticeTop ) 
    {
        new_value = LatticeTop;
    }
    else
    {
        if( old_value == LatticeTop )
        {
            new_value = LatticeTop;
        }
        else if( old_value == LatticeBot )
        {
            new_value = value;
        }
        else if( old_value == value )
        {
            new_value = old_value;
        }
        else
        {
            new_value = LatticeTop;
        }
    }
    
    if( new_value != old_value )
    {
        ABBL_EVAL_IN_VALUE(ibbl) = new_value;
        ABBL_FLAGS(ibbl) |= F_MARKED;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblInitEval"

static void AbblInitEval(INDEX ibbl,REG reg, INT64 LatticeBot, INT64 LatticeTop)
{
    INDEX iins;
    INDEX ifun = ABBL_FUN(ibbl);
    INT64 value;

    CASSERT( REG2BIT(reg) & REGS_ALL_BUT_ZERO,("bad register: %d\n",reg));
    
    if( AFUN_IS_PSEUDO(ifun) )
    {
        ABBL_EVAL_IN_VALUE(ibbl) = LatticeTop;
        ABBL_EVAL_DEF_VALUE(ibbl) = LatticeTop;                
        ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = LatticeTop;
        return;
    }

    ABBL_EVAL_IN_VALUE(ibbl) = LatticeBot;
    ABBL_EVAL_DEF_VALUE(ibbl) = LatticeBot;
    ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = LatticeBot;

    FOREACH_BBL_INS_R(ibbl,iins)
    {
        if( AINS_REG_C(iins) == reg  )
        {
            if( AINS_FLAGS(iins) & FI_EVAL_C )
            {
                value = AINS_C_EVAL(iins);                
                ABBL_EVAL_DEF_VALUE(ibbl) = value;
                ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = value;
                break;
            }
            else
            {
                ABBL_EVAL_DEF_VALUE(ibbl) = LatticeTop;
                ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = LatticeTop;
                break;
            }
        }
    }

    /* this is the new hack that propagates stuff over conditional */
    
    if( ABBL_TYPE(ibbl) == BT_BRANCH )
    {
        INDEX   iins_branch;
        REGLIST spoiled;
        INT64 value;
        
        iins_branch = ABBL_INS_LAST(ibbl);
        if( AINS_CODE(iins_branch) == I_bne || AINS_CODE(iins_branch) == I_beq )
        {
            spoiled = REGS_NONE;
            for(iins=AINS_PREV(iins_branch);VALID(iins);iins = AINS_PREV(iins))
            {
                spoiled |= AinsDef(iins);
                if( AINS_REG_C(iins) == AINS_REG_A(iins_branch) ) break;
            }

            if( reg  == AINS_REG_A(iins_branch) )
            {
                if( AINS_CODE(iins_branch) == I_beq )
                {
                    ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = 0;
                    if( VALID(iins) && AINS_TYPE(iins) == IT_ICP && AINS_CODE(iins) != I_cmpbge )
                        ABBL_EVAL_DEF_VALUE(ibbl) = 1;
                    return;
                }
                else 
                {
                    ABBL_EVAL_DEF_VALUE(ibbl) = 0;
                    if( VALID(iins) && AINS_TYPE(iins) == IT_ICP && AINS_CODE(iins) != I_cmpbge )
                        ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = 1;
                    return;
                }
            }

            if( VALID(iins) && !( REG2BIT(reg) & spoiled) )
            {
                if( reg == AINS_REG_A(iins) && (AINS_FLAGS(iins)&FI_EVAL_B) )
                    value = AINS_B_EVAL(iins);
                else if( reg == AINS_REG_B(iins) && (AINS_FLAGS(iins)&FI_EVAL_A) )
                    value = AINS_A_EVAL(iins);
                else
                {
                    return;
                }
#if 1
                switch( AINS_CODE(iins) )
                {
                case I_cmpeq:
                    if( AINS_CODE(iins_branch) == I_beq )
                        ABBL_EVAL_DEF_VALUE(ibbl) = value;
                    else
                        ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = value;
                    break;

                case I_addq:
                    if( AINS_CODE(iins_branch) == I_bne )
                        ABBL_EVAL_DEF_VALUE(ibbl) = -value;
                    else
                        ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = -value;
                    break;
                case I_subq:
                case I_xor:
                    if( AINS_CODE(iins_branch) == I_bne )
                        ABBL_EVAL_DEF_VALUE(ibbl) = value;
                    else
                        ABBL_EVAL_DEF_VALUE_TRUE(ibbl) = value;
                    break;

                }
#endif
            }
        }
    }
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "WarnAboutChange"

static void WarnAboutChange(INDEX ibbl, INDEX iins, REG reg)
{
#if 0    
    if( AINS_IS_LOAD(iins) && reg == REG_GP ) return;
    if( AbblNpreds(ibbl) == 0 ) return;
    if( AbblNpreds(ibbl) == 1 &&
        AEDG_HEAD(ABBL_PRED(ibbl)) == ibbl) return;
    
    ERROR(("reg %d known -> unknown\n",reg));
    PrintIns(stdout,iins);
    PrintBblSuper(stdout,AINS_BBL(iins)-1);
    PrintBblSuper(stdout,AINS_BBL(iins));
    PrintBblSuper(stdout,AINS_BBL(iins)+1);
#endif    
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblEvalPropagate"

static void AbblEvalPropagate(INDEX ibbl, REG reg, INT64 LatticeBot, INT64 LatticeTop)
{
    INT64 value = ABBL_EVAL_IN_VALUE(ibbl);
    INDEX iins;
    
    CASSERT( reg!=REG_ZERO && reg!=REG_FZERO && reg!=REG_LIT,("weird register %d in bbl: %d\n",reg,ibbl));

    FOREACH_BBL_INS(ibbl,iins)
    {
        if( AINS_REG_A(iins) == reg )
        {
            if( value != LatticeTop && value != LatticeBot )
            {
                AINS_A_EVAL(iins) = value;
                AINS_FLAGS(iins) |= FI_EVAL_A;
            }
            else
            {
                if( AINS_FLAGS(iins) & FI_EVAL_A )
                {
                    WarnAboutChange(ibbl,iins,reg);
                }
                AINS_FLAGS(iins) &= ~FI_EVAL_A;
            }
        }

        if( AINS_REG_B(iins) == reg )
        {
            if( value != LatticeTop && value != LatticeBot )
            {
                AINS_B_EVAL(iins) = value;
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }
            else
            {
                if( AINS_FLAGS(iins) & FI_EVAL_B )
                {
                    WarnAboutChange(ibbl,iins,reg);
                }
                AINS_FLAGS(iins) &= ~FI_EVAL_B;
            }
        }

        if( AINS_REG_C(iins) == reg  )
        {
            if( AINS_FLAGS(iins) & FI_EVAL_C )
            {
                value = AINS_C_EVAL(iins);
            }
            else
            {
                value = LatticeTop;
            }
        }
    }
}


/*=======================================================
=======================================================*/

#define CFA1

#undef __PROC__
#define __PROC__ "EvalConstPropagation"

static void EvalConstPropagation(REG reg, INT64 LatticeBot, INT64 LatticeTop)
{
    BOOL  flag;
    INT64 value;
    INT64 value_true;
    INDEX iins;
    INDEX ibbl;
    INDEX iedg;
    INDEX succ;
    INDEX succ_true;
    
#if 0
    printf("PROPAGATING REGISTER %d\n",reg);
#endif
    
    /* Initialize Fixpoint Computation */
    

    FOREACH_LIVE_BBL(ibbl)
    {
        AbblInitEval(ibbl,reg,LatticeBot,LatticeTop);
        ABBL_MARK(ibbl);
    }
        
    do
    {
        flag = 0;
        
        FOREACH_LIVE_BBL(ibbl)
        {
            if( !ABBL_IS_MARKED(ibbl) ) continue;
            
            ABBL_UNMARK(ibbl);
            flag = 1;


            /* COMPUTE OUT SET */

            value = ABBL_EVAL_DEF_VALUE(ibbl);
            if( value == LatticeBot )
            {
                value = ABBL_EVAL_IN_VALUE(ibbl);
            }

            value_true = ABBL_EVAL_DEF_VALUE_TRUE(ibbl);
            if( value_true == LatticeBot )
            {
                value_true = ABBL_EVAL_IN_VALUE(ibbl);
            }
            

            /* PROPAGATE OUT SET TO INSETS */

            if( ABBL_TYPE(ibbl) == BT_BRANCH )
            {
                succ_true = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
                succ = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));

                CASSERT( AbblNsuccs(ibbl) == 2,("weird\n") );
                                 
                EvalLatticJoin(succ,value,LatticeBot,LatticeTop);
                EvalLatticJoin(succ_true,value_true,LatticeBot,LatticeTop);
            }
            else
            {
                
                FOREACH_BBL_SUCC(ibbl,iedg)
                {
                    succ = AEDG_TAIL(iedg);
#if 0
                    /* this is just a minor speed optimization */

                    if( AFUN_IS_PSEUDO(ABBL_FUN(succ))  ) continue;
#endif
                
#ifdef CFA1
                    /* propagate unchanged regs */
                   
                    if( AEDG_TYPE(iedg) == ET_FUN_LINK )
                    {
                        INDEX CallFun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
                        REGLIST NotChanged = ~AFUN_REGS_CHANGED(CallFun);
                        if( ABBL_IS_CRIT_PATH(ibbl) || (REG2BIT(reg) & NotChanged) )
                            EvalLatticJoin(succ,value, LatticeBot, LatticeTop);
                        continue;
                    }

                    /* propagate changed regs only */
                
                    if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
                    {
                        INDEX CallFun = ABBL_FUN(ibbl);
                        REGLIST Changed = AFUN_REGS_CHANGED(CallFun);
                        if( ABBL_IS_CRIT_PATH(succ) || (REG2BIT(reg) & Changed) )
                            EvalLatticJoin(succ,value, LatticeBot, LatticeTop);
                        continue;
                    }
#else
                    if( AEDG_TYPE(iedg) == ET_FUN_LINK )
                    {
                        continue;
                    }
#endif                                        
                    EvalLatticJoin(succ,value, LatticeBot, LatticeTop);
                }
                
            }
            
        }
    }
    while( flag );

#if 1

    /* without this hack the code in fini and init sections used to break */
    /* this is no longer true but I left it anyway */

    if( reg == REG_PV )
    {
        FOREACH_INS(iins)
        {
            if( !(AINS_FLAGS(iins) & F_REFQUAD) ) continue;
            ibbl = AINS_BBL(iins);
            if( !ABBL_IS_FIRST(ibbl) ) continue;
            ABBL_EVAL_IN_VALUE(ibbl) = Ins2Addr(iins);
            /*
            printf("patching %s %16lx\n",AFUN_NAME(ABBL_FUN(ibbl)),Ins2Addr(iins));
            */
        }
    }
#endif

    /* update field */

    FOREACH_LIVE_BBL(ibbl)
    {
        AbblEvalPropagate(ibbl,reg,LatticeBot,LatticeTop);
    }

}


/*=======================================================
=======================================================*/

static void EvalComputeLatticeTopAndBot(FILE *fp)
{
    INDEX iins;

restart:

    FOREACH_INS(iins)
    {
        if( AINS_FLAGS(iins) & FI_EVAL_A )
        {
            if( AINS_A_EVAL(iins) == LatticeTop )
            {
                LatticeTop+=2;
                WARNING(("Changing top: %lx  bot: %lx\n",LatticeTop,LatticeBot));
                goto restart;
            }

            if( AINS_A_EVAL(iins) == LatticeBot )
            {
                LatticeBot+=2;
                WARNING(("Changing top: %lx  bot: %lx\n",LatticeTop,LatticeBot));
                goto restart;
            }
        }
        
        if( AINS_FLAGS(iins) & FI_EVAL_B )
        {
            if( AINS_B_EVAL(iins) == LatticeTop )
            {
                LatticeTop+=2;
                WARNING(("Changing top: %lx  bot: %lx\n",LatticeTop,LatticeBot));
                goto restart;
            }

            if( AINS_B_EVAL(iins) == LatticeBot )
            {
                LatticeBot+=2;
                WARNING(("Changing top: %lx  bot: %lx\n",LatticeTop,LatticeBot));
                goto restart;
            }
        }

        if( AINS_FLAGS(iins) & FI_EVAL_C )
        {
            if( AINS_C_EVAL(iins) == LatticeTop )
            {
                LatticeTop+=2;
                WARNING(("Changing top: %lx  bot: %lx\n",LatticeTop,LatticeBot));
                goto restart;
            }

            if( AINS_C_EVAL(iins) == LatticeBot )
            {
                LatticeBot+=2;
                WARNING(("Changing top: %lx  bot: %lx\n",LatticeTop,LatticeBot));
                goto restart;
            }
        }
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "OptPartialEvaluate"

extern BOOL OptPartialEvaluate(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Eval");
    BOOL   change = 0;
    NUMBER rcount=0;
    NUMBER ecount=0;
    NUMBER fecount=0;
    NUMBER scount=0;
    NUMBER fscount=0;
    INDEX  iins;
    REG    reg;
    REGLIST regs_changed;
    clock_t time = clock();
        
    
    EmulInit();

#if 1

    EvalComputeLatticeTopAndBot(fp);
    
    regs_changed = 0;
    FOREACH_LIVE_INS(iins)
    {
        if( AINS_FLAGS(iins) & FI_EVAL_C )
        {
            regs_changed |= REG2BIT(AINS_REG_C(iins));
        }
    }

    regs_changed &= ~(B_ZERO|B_FZERO);
    
    for(reg=0;reg<64;reg++)
    {
        if( REG2BIT(reg) & regs_changed )
        {
            EvalConstPropagation(reg,LatticeBot,LatticeTop);
        }
        
    }
#endif
        
    do
    {
        rcount++;
        regs_changed = 0;
        
        FOREACH_LIVE_INS(iins)
        {
            /*
             * Do straight forward partial evaluations
             */
            
            reg = AINS_REG_A(iins);
            
            if( reg == REG_LIT )
            {
                AINS_A_EVAL(iins) = AINS_DATA(iins);                    
                AINS_FLAGS(iins) |= FI_EVAL_A;
            }
            else if( reg == REG_ZERO || reg == REG_FZERO )
            {
                AINS_A_EVAL(iins) = 0;
                AINS_FLAGS(iins) |= FI_EVAL_A;
            }
            
            reg = AINS_REG_B(iins);
            
            if( reg == REG_LIT )
            {
                AINS_B_EVAL(iins) = AINS_DATA(iins);                    
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }
            else if( reg == REG_ZERO || reg == REG_FZERO )
            {
                AINS_B_EVAL(iins) = 0;
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }
            else if( reg == REG_GP )
            {
                if( AINS_TYPE(iins) == IT_ILD ||
                    AINS_TYPE(iins) == IT_LDA ||
                    AINS_TYPE(iins) == IT_FLD )
                {
                    AINS_B_EVAL(iins) = SysGetGlobalPointer();
                    AINS_FLAGS(iins) |= FI_EVAL_B;
                }
            }
            
            reg = AINS_REG_C(iins);  /* important for jmp and bra evaluation */

            if( reg == REG_ZERO || reg == REG_FZERO )
            {
                AINS_C_EVAL(iins) = 0;
                AINS_FLAGS(iins) |= FI_EVAL_C;
            }

            scount += RegisterForward(iins);

            /*
              fill in eval c - partial evaluation

              */


            if( !(AINS_FLAGS(iins) & FI_EVAL_C) )
            {

                switch( AINS_TYPE(iins) )
                {

                case IT_ULD:
                case IT_ILD:
                    ecount += EvalILD(iins);
                    break;

                case IT_JMP:
                case IT_BRA:
                    AINS_C_EVAL(iins) = Ins2Addr(iins+1);   /* NEEDS WORK */
                    AINS_FLAGS(iins) |= FI_EVAL_C;
                    ecount++;
                    break;

                case IT_LDA:
                    ecount += EvalLDA(iins);
                    break;
#if 1
                case IT_FCM:
                case IT_FCP:
                case IT_FOP:
                case IT_FM4:
                case IT_FM8:
                case IT_FD4:
                case IT_FD8:
                case IT_FCY:
                    fecount += EvalFOP(iins);
                    break;
#endif
                case IT_ICM:
                case IT_IOP:
                case IT_ICP:
                case IT_MLL:
                case IT_MLQ:
                case IT_MLH:
                case IT_EXT:
                case IT_LOG:
                case IT_SFT:                    
                    ecount += EvalIOP(iins);
                    break;
                    

                case IT_FLD:
                    fecount += EvalFLD(iins);
                    break;
                }

                if( AINS_FLAGS(iins) & FI_EVAL_C )
                {
                    regs_changed |= REG2BIT(AINS_REG_C(iins));
                }

            }

            /*
              strength reduction
              */
            
            if( !(AINS_FLAGS(iins)& F_KILLED) )
            switch( AINS_TYPE(iins) )
            {
            case IT_MLL:
            case IT_MLQ:
            case IT_MLH:
                scount += StrRedMUL(iins);
                break;
                
            case IT_ULD:
            case IT_UST:
                scount += StrRedUnaligned(iins);
                break;

            case IT_ICM:
                scount += StrRedICM(iins);
                break;

            case IT_FCM:
                fscount += StrRedFCM(iins);
                break;

            case IT_IBR:
                scount += StrRedIBR(iins);
                break;
                
            case IT_FBR:
                fscount += StrRedFBR(iins);
                break;

            case IT_JMP:
                scount += StrRedJMP(iins);
                break;

            }
        }

        if( verbose )
        {
            fprintf(fp,"\t[*] registers: %016lx Eval: %d+%d  Str-Red:: %d+%d\n",
                    regs_changed,ecount,fecount,scount,fscount);
        }

        EvalComputeLatticeTopAndBot(fp);
            
        for(reg=0;reg<64;reg++)
        {
            if( REG2BIT(reg) & regs_changed )
            {
                EvalConstPropagation(reg,LatticeBot,LatticeTop);
            }
            
        }
        
    }
    while( regs_changed );

    EmulExit();
    
    change = (ecount+fecount) > 0;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    

    InfoStatus(fp,"\t[OptEval%c] Eval: %d.%d  Str-Red:: %d.%d  (Rounds: %d) {%d ms}\n",
            CHANGE(change),ecount,fecount,scount,fscount,rcount,time);

    return change;
}

/*=======================================================
=======================================================*/


#undef __PROC__
#define __PROC__ "ProfPartialEvaluate"

extern BOOL ProfPartialEvaluate(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Eval");
    BOOL   change = 0;
    NUMBER rcount=0;
    NUMBER ecount=0;
    NUMBER fecount=0;
    NUMBER scount=0;
    NUMBER fscount=0;
    INDEX  iins;
    REG    reg;
    REGLIST regs_changed;
    clock_t time = clock();
        

    EmulInit();

#if 1
    regs_changed = 0;
    FOREACH_LIVE_INS(iins)
    {
        if( AINS_FLAGS(iins) & FI_EVAL_C )
        {
            regs_changed |= REG2BIT(AINS_REG_C(iins));
        }
    }

    regs_changed &= ~(B_ZERO|B_FZERO);

    EvalComputeLatticeTopAndBot(fp);
    
    for(reg=0;reg<64;reg++)
    {
        if( REG2BIT(reg) & regs_changed )
        {
            EvalConstPropagation(reg,LatticeBot,LatticeTop);
        }
        
    }
#endif
        
    do
    {
        rcount++;
        regs_changed = 0;
        
        FOREACH_LIVE_INS(iins)
        {
            /*
             * Do straight forward partial evaluations
             */
            
            reg = AINS_REG_A(iins);
            
            if( reg == REG_LIT )
            {
                AINS_A_EVAL(iins) = AINS_DATA(iins);                    
                AINS_FLAGS(iins) |= FI_EVAL_A;
            }
            else if( reg == REG_ZERO || reg == REG_FZERO )
            {
                AINS_A_EVAL(iins) = 0;
                AINS_FLAGS(iins) |= FI_EVAL_A;
            }
            
            reg = AINS_REG_B(iins);
            
            if( reg == REG_LIT )
            {
                AINS_B_EVAL(iins) = AINS_DATA(iins);                    
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }
            else if( reg == REG_ZERO || reg == REG_FZERO )
            {
                AINS_B_EVAL(iins) = 0;
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }
            else if( reg == REG_GP )
            {
                if( AINS_TYPE(iins) == IT_ILD ||
                    AINS_TYPE(iins) == IT_LDA ||
                    AINS_TYPE(iins) == IT_FLD )
                {
                    AINS_B_EVAL(iins) = SysGetGlobalPointer();
                    AINS_FLAGS(iins) |= FI_EVAL_B;
                }
            }
            
            reg = AINS_REG_C(iins);  /* important for jmp and bra evaluation */

            if( reg == REG_ZERO || reg == REG_FZERO )
            {
                AINS_C_EVAL(iins) = 0;
                AINS_FLAGS(iins) |= FI_EVAL_C;
            }

            scount += RegisterForward(iins);

            /*
              fill in eval c - partial evaluation

              */


            if( !(AINS_FLAGS(iins) & FI_EVAL_C) )
            {

                switch( AINS_TYPE(iins) )
                {

                case IT_ULD:
                case IT_ILD:
                    ecount += EvalILD(iins);
                    break;

                case IT_JMP:
                case IT_BRA:
                    AINS_C_EVAL(iins) = Ins2Addr(iins+1);   /* NEEDS WORK */
                    AINS_FLAGS(iins) |= FI_EVAL_C;
                    ecount++;
                    break;

                case IT_LDA:
                    ecount += EvalLDA(iins);
                    break;
#if 1
                case IT_FCM:
                case IT_FCP:
                case IT_FOP:
                case IT_FM4:
                case IT_FM8:
                case IT_FD4:
                case IT_FD8:
                case IT_FCY:
                    fecount += EvalFOP(iins);
                    break;
#endif
                case IT_ICM:
                case IT_IOP:
                case IT_ICP:
                case IT_MLL:
                case IT_MLQ:
                case IT_MLH:
                case IT_EXT:
                case IT_LOG:
                case IT_SFT:                    
                    ecount += EvalIOP(iins);
                    break;
                    

                case IT_FLD:
                    fecount += EvalFLD(iins);
                    break;
                }

                if( AINS_FLAGS(iins) & FI_EVAL_C )
                {
                    regs_changed |= REG2BIT(AINS_REG_C(iins));
                }

            }

            /*
              strength reduction
              */
#if 0
	    if (AINS_TYPE(iins) == IT_JMP) {
                scount += StrRedJMP(iins);
            }
#endif
        }

        if( verbose )
        {
            fprintf(fp,"\t[*] registers: %016lx Eval: %d+%d  Str-Red:: %d+%d\n",
                    regs_changed,ecount,fecount,scount,fscount);
        }

        EvalComputeLatticeTopAndBot(fp);
            
        for(reg=0;reg<64;reg++)
        {
            if( REG2BIT(reg) & regs_changed )
            {
                EvalConstPropagation(reg,LatticeBot,LatticeTop);
            }
            
        }
    }
    while( regs_changed );

    EmulExit();
    
    change = (ecount+fecount) > 0;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    

    InfoStatus(fp,"\t[OptProfEval%c] Eval: %d.%d  Str-Red:: %d.%d  (Rounds: %d) {%d ms}\n",
            CHANGE(change),ecount,fecount,scount,fscount,rcount,time);

        
    return change;
}


/*=======================================================
=======================================================*/


extern void ProfStrRedIndirectCalls()
{
    INDEX ibbl, iins;
    REG targetReg;
    ADDRESS addr;
    REG reg;
    INDEX jump;

    EmulInit();

    FOREACH_BBL(ibbl) {
	if (ABBL_TYPE(ibbl)==BT_CALL) {
	    addr = 0;
	    jump = ABBL_INS_LAST(ibbl);
	    /* can we figure out the target of this call? */
	    targetReg = AINS_REG_B(jump);
	    FOREACH_BBL_INS_R(ibbl, iins) {
		/* look for the last instruction that defines this reg */
		if( REG2BIT(targetReg) & AinsDef(iins) )  {
		    if( AINS_TYPE(iins) == IT_ILD ||
			AINS_TYPE(iins) == IT_LDA ||
			AINS_TYPE(iins) == IT_FLD )
		    {
			reg = AINS_REG_B(iins);
			if( reg == REG_GP ) {
			    AINS_B_EVAL(iins) = SysGetGlobalPointer();
			    AINS_FLAGS(iins) |= FI_EVAL_B;
			}
			
			/* ok, flags and b_eval field are set, now reduce this 
			   operation, and read out the eval_c field */
			
			EvalILD(iins);
			addr = AINS_C_EVAL(iins);
			break;
		    }
		}
	    }

	    /* if we have a valid address, fix the jump */
	    if (addr != 0) {
		AINS_B_EVAL(jump) = addr;
		AINS_FLAGS(jump) = FI_EVAL_B;
		StrRedJMP(jump);
	    }
	    
	}
    }

    EmulExit();    
}


/*=======================================================
=======================================================*/
