/*=======================================================
  Author:   R. Muth
  Version:  $Id: flowgraph.c,v 1.33 1998/10/06 18:02:37 muth Exp muth $
  Comment:  This file contains the functions related to the flowgraph
=======================================================*/

#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
=======================================================*/
#define FB_NOT_CRITICAL  (1<<5)
#define ABBL_IS_NOT_CRITICAL(i) (ABBL_FLAGS(i) & FB_NOT_CRITICAL)
#define FB_NOT_REACH_INIT (1<<6)
#define ABBL_IS_REACH_INIT(i) (ABBL_FLAGS(i) & FB_NOT_REACH_INIT)
#define FB_NOT_REACH_EXIT (1<<7)
#define ABBL_IS_REACH_EXIT(i) (ABBL_FLAGS(i) & FB_NOT_REACH_EXIT)
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "MarkFunctions"

extern void MarkFunctions(FILE *fp)
{
    INDEX iins;

    /* patch second to last instruction */

    AINS_FLAGS(CONFIG_UINS-1) |= FI_BBL_END | FI_FUN_END;
        
    /* patch entry instruction to make it permanent */

    iins = Addr2Ins(SysGetProgramEntry());
    AINS_FLAGS(iins) |= FI_BBL_BEG | FI_FUN_BEG;
    AINS_FLAGS(iins) |= F_REFQUAD;

    /* patch first instruction to make it permanent (necessary?)*/

    iins = 0;
    AINS_FLAGS(iins) |= FI_BBL_BEG | FI_FUN_BEG;
    AINS_FLAGS(iins) |= F_REFQUAD;

    /* create instruction at the end of text segment */

    iins = AinsNew();
    AINS_CODE(iins) = I_call_pal;
    AINS_REG_A(iins) = REG_ZERO;
    AINS_REG_B(iins) = REG_ZERO;
    AINS_REG_C(iins) = REG_ZERO;
    AINS_FLAGS(iins) |= F_REFQUAD | FI_LAST | F_KILLED;


    /*  find functions that are bsr targets */

    FOREACH_INS(iins)     
    {
        if( AINS_CODE(iins) == I_bsr )
        {
            INDEX it = iins + 1 + AINS_DATA(iins);
            AINS_FLAGS(it) |= FI_FUN_BEG | FI_BBL_BEG;
            if(it)
                AINS_FLAGS(it-1) |= FI_BBL_END | FI_FUN_END;
        }
    }

    InfoStatus(fp,"\t[MarkFunctions]\n");    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "MarkBblocks"

extern void MarkBblocks(FILE *fp)
{
    INDEX iins;

    /* Do the Branches */
    
    FOREACH_INS(iins)
    {
        INDEX target;
        
        switch( AINS_TYPE(iins) )
        {
        case IT_IBR:
        case IT_FBR:
        case IT_BRA:
            target = iins + 1 + AINS_DATA(iins);

            AINS_FLAGS(iins)   |= FI_BBL_END;
            AINS_FLAGS(iins+1) |= FI_BBL_BEG;
            AINS_FLAGS(target) |= FI_BBL_BEG;
            AINS_FLAGS(target-1) |= FI_BBL_END;
            break;
            
        case IT_JMP:

            AINS_FLAGS(iins)     |= FI_BBL_END;
            AINS_FLAGS(iins+1)   |= FI_BBL_BEG;
            break;

            /* isolate pals "halt" and "trap" */

        case IT_PAL:
#if 0
            if( AINS_DATA(iins) == 0 /* || AINS_DATA(iins) == 0xaa */ )
            {
                AINS_FLAGS(iins-1) |= FI_BBL_END;
                AINS_FLAGS(iins)   |= FI_BBL_BEG | FI_BBL_END;
                AINS_FLAGS(iins+1) |= FI_BBL_BEG;
            }
#else
            AINS_FLAGS(iins)     |= FI_BBL_END;
            AINS_FLAGS(iins+1)   |= FI_BBL_BEG;
#endif
            break;
        }
    }

    InfoStatus(fp,"\t[MarkBblocks]\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CreateFunctionsAndBblocks"

extern void CreateFunctionsAndBblocks(FILE *fp)
{
    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    NUMBER fcount = 0;
    NUMBER bcount = 0;
    clock_t time = clock();

    /*
      Create AFUN_PSEUDO_HELL fun for non resolved jsr's and jmp's
      there are also edge from HELL to all reloc bbls
      */
    
    ifun = AfunNewFancy();
    CFATAL(ifun!=AFUN_PSEUDO_HELL,("Assumption for PSEUDO_HELL fun violated: %d\n",ifun));
    AFUN_GP(ifun) = SysGetGlobalPointer();   /* information no used NEEDS WORK */
    AFUN_NAME(ifun) = "-HELL-";
    AFUN_FLAGS(ifun) |= FF_PSEUDO;
    
    ibbl = AFUN_BBL_FIRST(ifun);
    ABBL_FLAGS(ibbl) |= F_REFQUAD;  /* ensure that it will not be killed */

   /*
      Create AFUN_PSEUDO_PAL fun for non resolved jsr's and jmp's
      there are also edge from HELL to all reloc bbls
      */
    
    ifun = AfunNewFancy();
    CFATAL(ifun!=AFUN_PSEUDO_PAL,("Assumption for PSEUDO_PAL fun violated: %d\n",ifun));
    AFUN_GP(ifun) = SysGetGlobalPointer();   /* information no used NEEDS WORK */
    AFUN_NAME(ifun) = "-PAL-";
    AFUN_FLAGS(ifun) |= FF_PSEUDO;
        
    ibbl = AFUN_BBL_FIRST(ifun);
    ABBL_FLAGS(ibbl) |= F_REFQUAD;  /* ensure that it will not be killed */


    /*
      Create AFUN_PSEUDO_CALL fun for non resolved jsr's 
     */
    
    ifun = AfunNewFancy();
    CFATAL(ifun!=AFUN_PSEUDO_CALL,("Assumption for PSEUDO_PAL fun violated: %d\n",ifun));
    AFUN_GP(ifun) = SysGetGlobalPointer();   /* information no used NEEDS WORK */
    AFUN_NAME(ifun) = "-CALL-";
    AFUN_FLAGS(ifun) |= FF_PSEUDO;
        
    ibbl = AFUN_BBL_FIRST(ifun);
    ABBL_FLAGS(ibbl) |= F_REFQUAD;  /* ensure that it will not be killed */

    
    /* the order in the following loop is important */
    
    FOREACH_INS(iins)     
    {
        /*
         * FUN BEG
         */

        if( AINS_FLAGS(iins) & FI_LAST )
            continue;
        
        if( AINS_FLAGS(iins) & FI_FUN_BEG )
        {
            ifun = AfunNewFancy();

            fcount++;
            AFUN_GP(ifun) = SysGetGlobalPointer();   /* NEEDS WORK */
            AFUN_FLAGS(ifun) = AINS_FLAGS(iins) & (F_GPREL32|F_REFQUAD);
            
            AFUN_NAME(ifun) = FindSymbol( Ins2Addr(iins) );
        }

        /* BBL BEG */

                
        if( AINS_FLAGS(iins) & FI_BBL_BEG )
        {
            ibbl = AbblNew();
            bcount++;
            ABBL_FUN(ibbl) = ifun;
            /* ABBL_OLD_ADDR(ibbl)  = iins; */
            ABBL_FLAGS(ibbl)     = AINS_FLAGS(iins) & (F_GPREL32|F_REFQUAD);
        }

        AINS_BBL(iins) = ibbl;
        AinsAppend(iins);
        
        /* BBL END */

        if( AINS_FLAGS(iins) & FI_BBL_END )
        {
            AbblAppendFancy(ibbl);
        }
        
        /* FUN END */

        if( AINS_FLAGS(iins) & FI_FUN_END )
        {
            /* no special treatment */ 
        }
    }


    ibbl = FindTargetBbl(SysGetProgramEntry());
    ABBL_FLAGS(ibbl) |= FB_PROG_ENTRY;
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[CreateFunctionsAndBblocks] fun: %d bbl: %d {%d ms}\n",fcount,bcount,time);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "CreateEdges"

extern void CreateEdges(FILE *fp)
{
    INDEX ibbl;
    INDEX target;
    INDEX next;
    INDEX iins;
    INDEX ifun;
    
    FOREACH_BBL(ibbl)
    {
        iins = ABBL_INS_LAST(ibbl);
        ifun = ABBL_FUN(ibbl);
            
        /* at this point only exit blocks are empty */
        
        if( !VALID(iins) )
        {
            continue;
        }
        

        if( ABBL_IS_RELOC(ibbl) )         /* create dummy edge to reloc bbls */
        {
            if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
            target = AFUN_BBL_FIRST(AFUN_PSEUDO_HELL);
            AedgLinkFancy( AedgNewFancy(target,ibbl,ET_HELL) );
        }
        
        switch( AINS_TYPE(iins) )
        {
        case IT_BRA:    
            if( AINS_CODE(iins) == I_br )
            {
                target = AINS_BBL(iins + 1 + AINS_DATA(iins));
                next = AINS_BBL(iins+1);
                
                if( AINS_REG_C(iins) == REG_ZERO )    /* this is the common case, kill unconditional branch */
                {
                    ABBL_TYPE(ibbl) = BT_NORM;
                    AedgLinkFancy(AedgNewFancy(ibbl,target,ET_NORM));
                    AinsKill(iins);
                }
                else /* this is the uncommon case, DO NOT kill unconditional branch that saves address*/
                {
                    ABBL_TYPE(ibbl) = BT_EXTRA;
                    AedgLinkFancy(AedgNewFancy(ibbl,target,ET_EXTRA));
                    if( target != next )
                    {
                        ERROR(("unconditional br in bbl %d in fun %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,target);
                        
                    }
                }
            }
            else if( AINS_CODE(iins) == I_bsr )
            {
                ABBL_TYPE(ibbl) = BT_CALL;
                target = AINS_BBL(iins + 1 + AINS_DATA(iins));
                next = AINS_BBL(iins+1);
                AedgLinkFancy( AedgNewFancy(ibbl,next,ET_FUN_LINK) );
                AedgLinkFancy( AedgNewFancy(ibbl,target,ET_FUN_CALL) );
            }
            else
            {
                ERROR(("Weird\n"));
            }
            break;
            
        case IT_JMP:
            if( AINS_CODE(iins) == I_jsr ) /* create call to dummy function */
            {
                ABBL_TYPE(ibbl) = BT_CALL;
                next = AINS_BBL(iins+1);
                target = AFUN_BBL_FIRST(AFUN_PSEUDO_HELL);
                AedgLinkFancy( AedgNewFancy(ibbl,next,ET_FUN_LINK));
                AedgLinkFancy( AedgNewFancy(ibbl,target,ET_FUN_CALL));
            }
            else if( AINS_CODE(iins) == I_jmp )  /* create call to dummy bbl */
            {
                if( AINS_REG_C(iins) != REG_ZERO )
                {
                    ERROR(("non zero ret reg in jmp\n"));
                    PrintBblSuper(stdout,ibbl);
                }
                
                ABBL_TYPE(ibbl) = BT_JUMP;
                target = AFUN_BBL_FIRST(AFUN_PSEUDO_HELL);
                AedgLinkFancy( AedgNewFancy(ibbl,target,ET_JUMP ) );
            }
            else if( AINS_CODE(iins) == I_ret )
            {
                ABBL_TYPE(ibbl) = BT_RET;
                AedgLinkFancy( AedgNewFancy(ibbl,ibbl,ET_EXIT_LINK) );
            }
            else
            {
                ERROR(("weird\n"));
            }
            break;

        case IT_PAL:    /* pal zero = halt (no successor) */
            if( AINS_DATA(iins) == 0  )
            {
                ABBL_TYPE(ibbl) = BT_HALT;
            }
            else
            {
                /* the following code makes the exit function look nicer */
                INDEX prev;
#if 0
                ABBL_TYPE(ibbl) = BT_NORM;
                next = AINS_BBL(iins+1);
                AedgLinkFancy(AedgNewFancy(ibbl,next,ET_NORM));
#else
                prev = AINS_PREV(iins);
                if( AINS_DATA(iins) == 0x83 &&
                    VALID(prev) &&
                    (AINS_CODE(prev) == I_addq || AINS_CODE(prev) == I_bis) &&
                    AINS_REG_C(prev) == REG_V0 &&
                    AINS_REG_B(prev) == REG_LIT &&
                    AINS_REG_A(prev) == REG_ZERO &&
                    AINS_DATA(prev) == 1 ) 
                {
                    HURRAH(("syscall exit detected bbl %d in fun %d [%s]\n",ibbl,ABBL_FUN(ibbl),AFUN_NAME(ABBL_FUN(ibbl))));
                    /*
                    PrintBblSuper(stdout,ibbl);
                    PrintBblSuper(stdout,ibbl+1);
                    */
                    AinsMakeRet(ABBL_INS_LAST(ibbl+1),REG_RA);
                    /*
                    PrintBblSuper(stdout,ibbl+1);
                    */
                }

                
                ABBL_TYPE(ibbl) = BT_PAL;
                next = ibbl+1;
                target = AFUN_BBL_FIRST(AFUN_PSEUDO_PAL);
                AedgLinkFancy( AedgNewFancy(ibbl,next,ET_FUN_LINK));
                AedgLinkFancy( AedgNewFancy(ibbl,target,ET_FUN_CALL));
#endif
            }
            break;
            
        case IT_IBR:
        case IT_FBR:
            ABBL_TYPE(ibbl) = BT_BRANCH;
            target = AINS_BBL(iins + 1 + AINS_DATA(iins));
            next = AINS_BBL(iins+1);
            AedgLinkFancy(AedgNewFancy(ibbl,target,ET_TRUE));
            AedgLinkFancy(AedgNewFancy(ibbl,next,ET_FALSE));
            break;

        default:   /* fall through */
            ABBL_TYPE(ibbl) = BT_NORM;
            next = AINS_BBL(iins+1);
            if( !(AINS_FLAGS(iins+1) & FI_LAST) )
                AedgLinkFancy(AedgNewFancy(ibbl,next,ET_NORM));
            break;
        }
    }

    InfoStatus(fp,"\t[CreateEdges]\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComEdgeCounts"

extern void ComputeEdgeCounts(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Edgecount");
     INDEX  iedg;
    INDEX  ibbl;
    BOOL   flag;
    NUMBER ecount = 0;
    NUMBER tcount = 0;
    NUMBER gcount = 0;
    INT64 count;

    if( !CONFIG_USE_PROFILE )
    {
        WARNING(("Computation of edge probabilities without profiling info skipped\n"));
        return;
    }
            
    /* initial edge probabilities */
    
    FOREACH_LIVE_EDG(iedg)
    {
        INDEX hd = AEDG_HEAD(iedg);
        INDEX tl = AEDG_TAIL(iedg);

        if( ABBL_FUN(hd) == AFUN_PSEUDO_HELL || ABBL_FUN(tl) == AFUN_PSEUDO_HELL )
        {
            AEDG_COUNT(iedg) = 0;
            continue;
        }

        if( ABBL_COUNT(hd) == 0 || ABBL_COUNT(tl) == 0 )
        {
            AEDG_COUNT(iedg) = 0;
            continue;
        }

        switch( AEDG_TYPE(iedg) )
        {
        case ET_NORM:
        case ET_EXTRA:
        case ET_UNCOND:
        case ET_EXIT_LINK:
        case ET_FUN_CALL:
            if( !VALID(ABBL_COUNT(hd)) )
            {
#if 0                
                ERROR(("bad count\n"));            
                PrintEdgSuper(stdout,iedg);
#endif                
            }
            
            AEDG_COUNT(iedg) = ABBL_COUNT(hd);
            break;
            
        case ET_FUN_RETURN:
            AEDG_COUNT(iedg) = AEDG_COUNT(iedg-1);
            break;

        case ET_SWITCH:
        case ET_TRUE:
        case ET_FALSE:
            AEDG_COUNT(iedg) = INVALID;
            break;
            
        case ET_JUMP:
        case ET_HELL:
        case ET_COMPENSATE:
        case ET_SETLONGJMP:
        case ET_FUN_LINK:
            AEDG_COUNT(iedg) = 0;
            break;
            
        default:
            ERROR(("unexpected edge type\n"));
            PrintEdgSuper(stdout,iedg);
            break;
        }
    }
    

    /* do as many edges as we can exactly! */
    do
    {
        flag = 0;

        FOREACH_LIVE_EDG(iedg)
        {
            INDEX hd;
            INDEX tl;
            INDEX jedg;

            tcount++;
            
            if( VALID(AEDG_COUNT(iedg)) ) continue;

            /*
              see whether we can compute edge count by
              looking at other outgoing edges
              */
               
            hd = AEDG_HEAD(iedg);
            count = ABBL_COUNT(hd);
            
            FOREACH_BBL_SUCC(hd,jedg)
            {
                if( jedg == iedg ) continue;
                if( !VALID(AEDG_COUNT(jedg)) )
                    break;
                else
                    count -= AEDG_COUNT(jedg);
            }
            
            if( !VALID(jedg) )
            {
                if( count < 0 )
                    count = 0;
                AEDG_COUNT(iedg) = count;
                flag = 1;
                ecount++;
                continue;
            }

            /*
              see whether we can compute edge count by
              looking at other outgoing edges
              */
            
            tl = AEDG_TAIL(iedg);
            count = ABBL_COUNT(tl);

            FOREACH_BBL_PRED(tl,jedg)
            {
                if( jedg == iedg )
                    continue;
                if( !VALID(AEDG_COUNT(jedg)) )
                    break;
                else
                    count -= AEDG_COUNT(jedg);
            }

            if( !VALID(jedg) )
            {
                if( count < 0 )
                    count = 0;
                AEDG_COUNT(iedg) = count;
                flag = 1;
                ecount++;
                continue;
            }
        }
    }
    while( flag );

    FOREACH_LIVE_BBL(ibbl)
    {
        NUMBER invalid = 0;
        count = ABBL_COUNT(ibbl);

        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
            
        FOREACH_BBL_SUCC(ibbl,iedg)
        {
            if( !VALID(AEDG_COUNT(iedg)) )
                invalid++;
            else
                count -= AEDG_COUNT(iedg);
        }

        if( invalid == 0 ) continue;

        gcount += invalid;
        if( count < 0)
            count = 0;
        count /= invalid;
        
        FOREACH_BBL_SUCC(ibbl,iedg)
        {
            if( VALID(AEDG_COUNT(iedg)) ) continue;
            if( verbose )
            {
                fprintf(fp,"\t[*] Guessing edge %d (%d->%d)\n",iedg,AEDG_HEAD(iedg),AEDG_TAIL(iedg));
            }
                
            AEDG_COUNT(iedg) = count;
        }
    }

    /* update ET_FUN_LINK */
    
    FOREACH_LIVE_EDG(iedg)
    {
        INDEX iedg_call;
        
        if( AEDG_TYPE(iedg) != ET_FUN_LINK ) continue;
        CASSERT( AEDG_COUNT(iedg) == 0,("nonzero ET_FUN_LINK count\n"));
        iedg_call = ABBL_EDG_CALL(AEDG_HEAD(iedg));
        CASSERT( AEDG_TYPE(iedg_call) == ET_FUN_CALL,("non ET_FUN_CALL edge\n"));
        AEDG_COUNT(iedg) = AEDG_COUNT(iedg_call);
    }
        
    InfoStatus(fp,"\t[ComputeEdgecount] total: %d  exact: %d  guessed: %d\n",
           tcount,ecount,gcount);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPropagateMarkForward"

static void AfunPropagateMarkForward(INDEX ifun,INDEX ibbl)
{
    INDEX iedg;
    
    if( ABBL_FLAGS(ibbl) & FB_NOT_CRITICAL ) return;
    if( ABBL_FLAGS(ibbl) & F_MARKED ) return;

    ABBL_FLAGS(ibbl) |= F_MARKED;

    FOREACH_BBL_SUCC(ibbl,iedg)
    {
        if( AEDG_IS_INTER(iedg) ) continue;
        AfunPropagateMarkForward(ifun,AEDG_TAIL(iedg));
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunReachInit"

static BOOL AfunReachInit(INDEX ifun)
{
    INDEX ibbl;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~F_MARKED;
    }

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        if( ABBL_HAS_PREDS(ibbl) ) continue;
        
        ABBL_FLAGS(ibbl) |= F_MARKED;
    }
    
    AfunPropagateMarkForward(ifun, AFUN_BBL_FIRST(ifun));

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        if( !ABBL_IS_REACH_INIT(ibbl) ) continue;
        if( ABBL_FLAGS(ibbl) & F_MARKED ) continue;
        return 0;
    }

    return 1;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPropagateMarkBackward"

static void AfunPropagateMarkBackward(ifun,ibbl)
{
    INDEX iedg;
    
    if( ABBL_FLAGS(ibbl) & FB_NOT_CRITICAL ) return;
    if( ABBL_FLAGS(ibbl) & F_MARKED ) return;

    ABBL_FLAGS(ibbl) |= F_MARKED;

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        if( AEDG_IS_INTER(iedg) ) continue;
        AfunPropagateMarkBackward(ifun,AEDG_HEAD(iedg));
    }
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunReachInit"

static BOOL AfunReachExit(INDEX ifun)
{
    INDEX ibbl;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~F_MARKED;
    }

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        if( ABBL_HAS_SUCCS(ibbl) ) continue;
        
        ABBL_FLAGS(ibbl) |= F_MARKED;
    }

    AfunPropagateMarkBackward(ifun, AFUN_BBL_LAST(ifun));

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) ) continue;
        if( !ABBL_IS_REACH_EXIT(ibbl) ) continue;
        if( ABBL_FLAGS(ibbl) & F_MARKED ) continue;
        return 0;
    }

    return 1;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunComputeCriticalBbls"

extern void AfunComputeCriticalBbls(INDEX ifun)
{
    INDEX ibbl;
    BOOL   flag;
    INT64  count;

    
    /* INITIALIZE SEEDS */

    flag = 0;
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~(FB_CRITICAL | FB_NOT_CRITICAL);
    
        if( ABBL_IS_HOT(ibbl) )
        {
            ABBL_FLAGS(ibbl) |= FB_CRITICAL;
            flag=1;
        }

        if( ABBL_COUNT(ibbl) <= 0 )
        {
            ABBL_FLAGS(ibbl) |= FB_NOT_CRITICAL;
        }
        
    }

    if( flag == 0)
        return;

    /* MARK UNREACHABLE BBLS - INIT*/

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~(F_MARKED|FB_NOT_REACH_INIT);
    }

    AfunPropagateMarkForward(ifun,AFUN_BBL_FIRST(ifun));

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_FLAGS(ibbl) & F_MARKED )
        {
            ABBL_FLAGS(ibbl) |= FB_NOT_REACH_INIT;
        }
    }

    /* MARK UNREACHABLE BBLS - EXIT*/

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~(F_MARKED|FB_NOT_REACH_EXIT);
    }

    AfunPropagateMarkBackward(ifun,AFUN_BBL_LAST(ifun));

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_FLAGS(ibbl) & F_MARKED )
        {
            ABBL_FLAGS(ibbl) |= FB_NOT_REACH_EXIT;
        }
    }

    ABBL_FLAGS(AFUN_BBL_FIRST(ifun)) |= FB_CRITICAL;
    ABBL_FLAGS(AFUN_BBL_LAST(ifun)) |= FB_CRITICAL;


    for( count=1;count<63;count++)
    {
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            if( ABBL_IS_CRITICAL(ibbl) )
            {
                if( ABBL_TYPE(ibbl) == BT_CALL )
                {
                    ABBL_FLAGS(AEDG_TAIL(ABBL_EDG_LINK(ibbl))) |= FB_CRITICAL;
                }
                continue;
            }
            if( ABBL_IS_NOT_CRITICAL(ibbl) ) continue;
            if( ABBL_COUNT(ibbl) > (1L << count) ) continue;
        
            ABBL_FLAGS(ibbl) |= FB_NOT_CRITICAL;
            if( AfunReachInit(ifun) && AfunReachExit(ifun) ) continue;
            CWARNING( ABBL_COUNT(ibbl) <= 0,("bad bbl count: %d\n",ABBL_COUNT(ibbl)));
        
            ABBL_FLAGS(ibbl) &= ~FB_NOT_CRITICAL;
            ABBL_FLAGS(ibbl) |= FB_CRITICAL;
        }
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeCriticalBbls"

extern void ComputeCriticalBbls(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Critical");
    INDEX ifun;
    INDEX ibbl;
    NUMBER bccount;
    NUMBER btcount;
    NUMBER bhcount;
    NUMBER ftcount;
    NUMBER itcount;
    NUMBER iccount;
    NUMBER ihcount;
    INT64  tweight;
    INT64  fweight;
    clock_t time = clock();
    
    if( !CONFIG_USE_PROFILE )
    {
        WARNING(("Computation of critical bbls without profiling info skipped\n"));
        return;
    }
    
    /* initialize some bbls to be on the critical path */

    FOREACH_LIVE_FUN(ifun)
    {
        if( AFUN_IS_PSEUDO(ifun) ) continue;

        AfunComputeCriticalBbls(ifun);
    }

    tweight = 0;
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_COUNT(ibbl) > 0 )
            tweight += ABBL_COUNT(ibbl) * AbblNInss(ibbl);
    }
    
    
    ftcount = 0;
    
    bhcount = 0;
    bccount = 0;
    btcount = 0;

    ihcount = 0;
    iccount = 0;
    itcount = 0;

    
    FOREACH_LIVE_FUN(ifun)
    {
        if( AfunNBblsHot(ifun) == 0 ) continue;

        ftcount++;
        fweight = 0;
        
        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            NUMBER n = AbblNInss(ibbl);

            btcount++;
            itcount+=n;

            if( ABBL_COUNT(ibbl) > 0 )
                fweight += ABBL_COUNT(ibbl) * n;
                    
            if( ABBL_IS_HOT(ibbl) )
            {
                bhcount++;
                ihcount+=n;
            }
            
            if( ABBL_IS_CRITICAL(ibbl) )
            {
                bccount++;
                iccount+=n;
            }
            
        }

        if( verbose )
        {
#if 0
            PrintFun(stdout,ifun);
#endif            
            printf("\t[*] %.4f %c%c%c%c%c%c%c%c %d [%s]  B: %d/%d  I: %d/%d  C: %d/%d\n",
                   (double)fweight/(double)tweight,
                   AfunIsLeaf(ifun)              ? 'L' : ' ',
                   AfunHasEsc2(ifun)             ? 'E' : ' ',
                   AfunHasEsc2Critical(ifun)     ? 'e' : ' ',
                   AfunHas2Esc(ifun)             ? 'I' : ' ',
                   AfunNSwitches(ifun)>0         ? 'S' : ' ',
                   AfunNSwitchesCritical(ifun)>0 ? 's' : ' ',
                   AfunNJumps(ifun)>0            ? 'J' : ' ',
                   AfunNJumpsCritical(ifun)>0    ? 'J' : ' ',
                   ifun,AFUN_NAME(ifun),
                   AfunNBblsCritical(ifun),AfunNBbls(ifun),
                   AfunNInssCritical(ifun),AfunNInss(ifun),
                   AfunNHotCallSites(ifun),AfunNCallSites(ifun));
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    InfoStatus(fp,"\t[ComputeCritical] Funs: %d  Bbls: %d/%d/%d  Inss: %d/%d/%d {%d ms}\n",
           ftcount,bhcount,bccount,btcount,ihcount,iccount,itcount,time);
}

/*=======================================================
=======================================================*/





