/*
  mset.h - prototypes and defines for implementation of mset adt.
  */
#ifndef _MSET_H_
#define _MSET_H_
#ifdef MSET_TESTING
typedef unsigned char  UINT8;
typedef   signed char  INT8;
typedef unsigned short UINT16;
typedef   signed short INT16;
typedef unsigned int   UINT32;
typedef   signed int   INT32;
typedef unsigned long  UINT64;
typedef   signed long  INT64;
typedef INT8  BOOL;
typedef INT32 INDEX;
typedef INT32 NUMBER;
typedef INT64 ADDRESS;
typedef char *STRING;
typedef UINT64 REGLIST;
typedef INT8  REG;
typedef UINT16 FLAGS;
typedef INT8  DATATYPE;

#ifndef FALSE
#define FALSE '\0'
#define TRUE '\1'
#endif /* ifndef FALSE */

#define INVALID -1

#else
#include "global.h"
#endif

#define K 6
#define MODULUS (1 << K)
#define SET_SIZE_BITS (1 << K)  /* 2^k */
#define MOD_MASK (SET_SIZE_BITS - 1)
#define SET_SIZE_LWORDS (1 << (K-6))

#define MODK(x) ((x)&MOD_MASK)

typedef unsigned int mset_value_t;

#define BITV_ALL   0xffffffffffffffff
#define BITV_EMPTY 0x0000000000000000

#define BITV_SINGLETON(v) (1UL <<(v))

#define BITV_UNION(s1,s2) ((s1) | (s2))
#define BITV_INTERSECT(s1,s2) ((s1) & (s2))
#define BITV_COMPL(s) (~(s))
#define BITV_DIFF(s1,s2) BITV_INTERSECT((s1),BITV_COMPL(s2))
#define BITV_ADD(s, v) BITV_UNION(BITV_SINGLETON(v),(s))

#define BITV_IS_EMPTY(rs) ((rs) == 0UL)
#define BITV_IS_SUBSET(subrs,rs) ((subrs) == ((subrs) & (rs)))
#define BITV_IS_MEMBER(member,rs) BITV_IS_SUBSET(BITV_SINGLETON(member),(rs))

#define BITV_FOREACH(regset,reg)   \
    for( reg = 0; reg <= 63; reg++)       \
      if( ! BITV_IS_MEMBER(reg,regset) ) \
         continue; else

#define BASE_TOP -2
#define BASE_BOT -1
#define VALID_BASE(i) (i >= 0)

#define MSET_IS_TOP(m) ((m)->set == BITV_ALL && (m)->basep == BASE_TOP)
#define MSET_IS_BOT(m) ((m)->set == BITV_EMPTY && (m)->basep == BASE_BOT)
	 
/* maintains the state for each basic block */
typedef struct mset {
  INDEX basep; /* index to instruction that defines the base value */
  UINT64 set;  /* values in the set */
} mset;

#ifndef MSET_TESTING
BOOL mset_sameValue(mset *m1, INDEX i1, mset *m2, INDEX i2);
BOOL mset_build(mset *m, INDEX instr, REG r);
#endif /* #ifndef MSET_TESTING */

void mset_all(mset *m);
void mset_empty(mset *d);
BOOL mset_union(mset *d, mset *m1, mset *m2);
void mset_singleton(mset *m, UINT64 v);
void mset_base(mset *m, INDEX bp);
BOOL mset_copy(mset *d, mset *s);

BOOL mset_disjoint(mset *m1, mset *m2);
BOOL mset_equal(mset *m1, mset *m2);

void mset_addi(mset *d, mset *s, UINT64 v, INDEX i);
void mset_add(mset *d, mset *m1, mset *m2, INDEX i);
void mset_subi(mset *d, mset *s, UINT64 v, INDEX i);
void mset_sub(mset *d, mset *m1, mset *m2, INDEX i);
void mset_muli(mset *d, mset *s, UINT64 v, INDEX i);
void mset_mul(mset *d, mset *m1, mset *m2, INDEX i);
void mset_printem(FILE *fp, mset *s);
#endif /* #ifndef MSET_H_ */
