/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.dupfun.c,v 1.1 1998/04/30 23:41:13 muth Exp muth $
  Comment:  This file contains the code for function duplication
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"


/*=======================================================
=======================================================*/
static BOOL AfunHasBadCallsite(INDEX ifun)
{
    INDEX iedg;
    INDEX init = AFUN_BBL_FIRST(ifun);

    FOREACH_BBL_PRED(init,iedg)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(AEDG_HEAD(iedg))) ) return 1;
    }

    return 0;
}
/*=======================================================
=======================================================*/
static BOOL AfunHasGoodCallsite(INDEX ifun)
{
    INDEX iedg;
    INDEX init = AFUN_BBL_FIRST(ifun);

    FOREACH_BBL_PRED(init,iedg)
    {
        if( !AFUN_IS_PSEUDO(ABBL_FUN(AEDG_HEAD(iedg))) ) return 1;
    }

    return 0;
}

/*=======================================================
@OPT
Duplicate Functions.
To reduce "bad data flow" into function called from unknown callsites,
eg. "hell".
All functions which have "normal" callsite and are called by hell
are cloned (duplicated). The normal callsites will call one clone
and hell will call the other.
It might be important that the clone which gets the flow from hell
etc has the higher function index so that the layouter moves
the code farther to the end
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptDuplicateFunctions"     

extern BOOL OptDuplicateFunctions(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Dupfun");
    BOOL   change;
    NUMBER fcount=0;
    NUMBER bcount=0;
    NUMBER icount=0;
    INDEX  init;
    INDEX callsite;
    INDEX callee;
    INDEX caller;
    INDEX ifun;
    INDEX iedg;
    double flow_frac;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN_R(callee)  /* reverse call is not really necessary */
    {

        if( AFUN_IS_PSEUDO(callee) ||
            AfunIsSetjmp(callee) ||
            AfunIsLongjmp(callee) ||
            AfunHasEsc2(callee) ||
            AfunNJumps(callee) > 0 )
        {
            if( verbose )
            {
                printf("\t[*] skipping bad function: %d [%s] (%ld)\n",callee,AFUN_NAME(callee),ABBL_COUNT(AFUN_BBL_FIRST(callee)));
            }
            continue;
        }

        if( !AfunHasBadCallsite(callee) ||
            !AfunHasGoodCallsite(callee) )
        {
            if( verbose )
            {
                printf("\t[*] skipping nodup function: %d [%s] (ff %ld)\n",callee,AFUN_NAME(callee),ABBL_COUNT(AFUN_BBL_FIRST(callee)));
            }
            continue;
        }
        

#if 0
        if( !VALID(AfunStackChange(callee)) ) continue;

        if( callee > PRINT_OPT_NUM ) continue;
#endif        
           
        init = AFUN_BBL_FIRST(callee);

        /*
         flow frac computation for the bad callsites
         */
        
        flow_frac = (double) ABBL_COUNT(init);

        FOREACH_BBL_PRED(init,iedg)
        {
            callsite = AEDG_HEAD(iedg);
            caller = ABBL_FUN(callsite);

            if( AEDG_TYPE(iedg) == ET_HELL ) continue;
            
            if( AEDG_TYPE(iedg) != ET_FUN_CALL )
            {
                WARNING(("unexpected %c edge to init bbl of fun %d [%s]\n",AEDG_TYPE(iedg),callee,AFUN_NAME(callee)));
                continue;
            }
            
            if( AFUN_IS_PSEUDO(caller) ) continue;
            
            if( ABBL_COUNT(callsite) > 0 )
                flow_frac -= (double) ABBL_COUNT(callsite);
        }

        if( flow_frac <= 1.0 )  /* this is to make up for some imprecission in pixie */
            flow_frac = 0.0;
        
        flow_frac /= 0.0001 + (double) ABBL_COUNT(init);

        ifun = AfunDup(callee,flow_frac); /* we take all the flow with us */
        
        bcount+=AfunNBbls(ifun);
        icount+=AfunNInss(ifun);
        fcount+=1;

        if( verbose )
        {
            printf("\t[*] duplicating: %d [%s] copy: %d (%g)\n",callee,AFUN_NAME(callee),ifun,flow_frac);
        }

        FOREACH_BBL_PRED(init,iedg)
        {
            callsite = AEDG_HEAD(iedg);
            caller = ABBL_FUN(callsite);

            if( AEDG_TYPE(iedg) != ET_HELL ) continue;
            CASSERT( AFUN_IS_PSEUDO(caller),("caller %d [%s] is not pseudo\n",caller,AFUN_NAME(caller)));
            CASSERT( ABBL_IS_RELOC(AFUN_BBL_FIRST(callee)),("non reloc bbl\n"));
            AbblMigrateRelocInfo(AFUN_BBL_FIRST(callee),AFUN_BBL_FIRST(ifun));
            AedgKillFancy(iedg);                     
            AedgLinkFancy( AedgNewFancy(callsite,AFUN_BBL_FIRST(ifun),ET_HELL) );
        }
        
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (fcount>0);

    InfoStatus(fp,"\t[OptDupfun%c] Duplicated Funs: %d  Bbls: %d  Inss: %d\n",
           CHANGE(change),fcount,bcount,icount);

    return change;
}

/*=======================================================
=======================================================*/

