#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
@OPT
Avoid Unneccessary Callee Register Savings
Not implemented yet but code is scanning for opportunities
and prints them.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptStackSaveRegs"

extern BOOL OptStackSaveRegs(FILE *fp)    
{
#if 1
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Stacksave");
#endif
    BOOL   change;
    NUMBER count=0;
    INDEX  ifun;
#if 0
    REG    reg;
#endif
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        REGLIST toomuch;

        if( AFUN_IS_PSEUDO(ifun) ) continue;
        
        toomuch = AFUN_REGS_SAVED(ifun) & ~AFUN_REGS_CHANGED(ifun);

        if( toomuch )
        {
            if( verbose )
            {
                fprintf(fp,"\t[*] s: %016lx c: %016lx - %016lx fun: %d [%s]\n",
                        AFUN_REGS_SAVED(ifun), AFUN_REGS_CHANGED(ifun), toomuch,
                        ifun,AFUN_NAME(ifun));
                
            }

#if 0            
            for( reg=0; reg<64; reg++ )
            {
                if( REG2BIT(reg) & toomuch )
                {
                    AfunKillLoadStoreSaveRegister(ifun,reg);
                    count++;
                }
                
            }
#endif            
            
        }
        
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (count>0);

    InfoStatus(fp,"\t[OptStacksave%c] Removed Stk Saves: %d {%d ms}\n",CHANGE(change),count,time);
        
    return change;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunCallsStackArgFuns"

extern BOOL AfunCallsStackArgFuns(INDEX ifun)    
{
    INDEX ibbl;
    INDEX callee;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        callee = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
        if( AFUN_HAS_STK_ARG(callee) )
        {
            return 1;
        }
        
    }
    return 0;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunUsesStack"

static BOOL AfunUsesStack(INDEX ifun)    
{
    INDEX ibbl;
    INDEX iins;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            /* no stack involved */
            
            if( AINS_REG_A(iins) != REG_SP && AINS_REG_B(iins) != REG_SP && AINS_REG_C(iins) != REG_SP )
                continue;
            
            /* lda is ok */
                
            if( AINS_REG_B(iins) == REG_SP && AINS_REG_C(iins) == REG_SP &&  AINS_TYPE(iins) == IT_LDA )
                continue;
            
            /* stores are ok too if not used */
            
            if( AINS_REG_B(iins) == REG_SP && (AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD) )
                continue;
            
            return 1;
        }
    }

    return 0;
}

/*=======================================================
@OPT
Delete Unused Stack.
A stackframe which is never used will be deleted.
This optimization seems to be  not very effective.
and prints them.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptDeleteStack"

extern BOOL OptDeleteStack(FILE *fp)    
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Stackdel");
    BOOL change = 0;
    INDEX ifun;
    INDEX iins;
    
    NUMBER count = 0;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        INDEX ibbl;

        /*
          do not try to remove stacks from weird functions
          */

        if( AFUN_IS_PSEUDO(ifun) ) continue;
        if( AfunIsSetjmp(ifun) ) continue;
        if( AfunIsLongjmp(ifun) ) continue;
        if( AFUN_HAS_STK_INVALID(ifun) ) continue;
        if( AFUN_STK_MIN(ifun) == 0 ) continue;
        if( AfunHas2Esc(ifun) ) continue;
        if( AfunHasEsc2(ifun) ) continue;
        if( VALID(AFUN_MASTER(ifun)) ) continue;
        if( AfunCallsStackArgFuns(ifun) ) continue;        
        if( AfunUsesStack(ifun) ) continue;
        
        count++;
                
        if( verbose )
        {
            InfoExtra(fp,"\t[*] Removing unused stack for %d [%s]\n",ifun,AFUN_NAME(ifun));
        }
        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_REG_A(iins) != REG_SP &&
                    AINS_REG_B(iins) != REG_SP &&
                    AINS_REG_C(iins) != REG_SP )
                    continue;

                AinsKill(iins);
            }
        }


    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (count>0);
    InfoStatus(fp,"\t[OptStackdel%c] Removed Stacks: %d {%d ms}\n",CHANGE(change),count,time);
    return change;
}
    
/*=======================================================
@OPT
Avoid Callee Save Registers in Leaf Functions
CURRENTLY NOT USED.
@END
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "OptLeafSaveRegisters"

extern BOOL OptLeafSaveRegisters(FILE *fp)    
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Leafsave");
    BOOL change = 0;
    INDEX ifun;
    NUMBER rcount = 0;
    NUMBER lcount = 0;
    REGLIST save;
    REGLIST scratch;
    REG reg_saved;
    REG reg_bypass;
    REG reg_return;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        if( AFUN_IS_PSEUDO(ifun) ) continue;
        if( !AfunIsLeaf(ifun) ) continue;
        if( AfunStackFrameSize(ifun) == 0 ) continue;
        lcount++;
        if( AfunHasEsc2(ifun) ) continue;
        if( AfunHas2Esc(ifun) ) continue;

        reg_return = AfunReturnRegister(ifun);
        CASSERT( VALID(reg_return),("invalid return reg\n"));
        
        save = AfunComputeSavedRegisters(ifun);
        save &= ~REG2BIT(reg_return);
        scratch = AfunScratchRegisters(ifun);
#if 0
        printf("%d [%s] saved: %016lx scratch: %016lx stk: %d\n",
               ifun,AFUN_NAME(ifun),save,scratch,AfunStackFrameSize(ifun));
#endif
        if( save == NO_REGS ) continue;
        if( scratch == NO_REGS ) continue;

#if 0        
        PrintFunSuper(stdout,ifun);
#endif        
        for( reg_saved = 0; reg_saved < 31; reg_saved++ )
        {
            if( !(save & REG2BIT(reg_saved)) ) continue;
            if( !(scratch & INT_REGS) ) continue;
            
            for( reg_bypass = 0; reg_bypass < 31; reg_bypass++ )
            {
                if( scratch & REG2BIT(reg_bypass) ) break;
            }

            scratch &= ~REG2BIT(reg_bypass);
            AfunBypassSaveRegister(ifun,reg_saved,reg_bypass);
            rcount++;
        }

        for( reg_saved = 32; reg_saved < 63; reg_saved++ )
        {
            if( !(save & REG2BIT(reg_saved)) ) continue;
            if( !(scratch & FLT_REGS) ) continue;
            
            for( reg_bypass = 32; reg_bypass < 63; reg_bypass++ )
            {
                if( scratch & REG2BIT(reg_bypass) ) break;
            }

            scratch &= ~REG2BIT(reg_bypass);
            AfunBypassSaveRegister(ifun,reg_saved,reg_bypass);
            rcount++;
        }
#if 0
        PrintFunSuper(stdout,ifun);
#endif        
    }
    

    InfoStatus(fp,"\t[OptLeafsave] Unsaved Regs: %d  (stk-leaf-funs: %d)\n",rcount,lcount);
        
    return (rcount > 0);
}
#endif    
/*=======================================================
@OPT
Delete unused stacks.
Currently almost identical to OptStackSaveRegs.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCalleeSaveRegisters"

extern BOOL OptCalleeSaveRegisters(FILE *fp)    
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Calleesave");
    BOOL change = 0;
    INDEX ifun;
#if 0
    INDEX ibbl;
    NUMBER rcount = 0;
    NUMBER lcount = 0;
#endif 
    REGLIST toomuch;
    INDEX exit;
    REG reg;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        if( AFUN_IS_PSEUDO(ifun) ) continue;
        AfunComputeSavedRegisters(ifun);
        exit = AFUN_BBL_LAST(ifun);

        toomuch = ~AbblRegsLiveIn(exit) & AFUN_REGS_SAVED(ifun);

        for( reg = 0; reg < 63; reg++ )
        {
            if( !(toomuch & REG2BIT(reg)) ) continue;
            if( verbose )
            {
                printf("\t[*] unnecessary register %d save in fun %d [%s]\n",reg,ifun,AFUN_NAME(ifun));
            }
            
        }
        
    }

            
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = 0;

    InfoStatus(fp,"\t[OptCalleesave%c] Unsaved Regs: {%d ms}\n",CHANGE(change),time);

    return change;
}


/*=======================================================
=======================================================*/
/*=======================================================
Stack merging Aaalysis by fixpoint computation


Lattic:
                      top
                     / | \
               concrete nat numbers
                     \ | /
                     bottom


Bottom is represented by -1 assuming that the stack change will
always be an even number.
Top is not represented because it we see top we abort the analysis.
                     

=======================================================*/

#define ABBL_STACK_CHANGE_IN(i) ABBL_U1(i).quadword
#define BOTTOM -1
#undef DEBUG_STK

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunComputeStackChange"

static void AfunComputeStackChange(INDEX ifun)
{
    BOOL repeat;
    INDEX ibbl;
    INDEX iins;
    INDEX succ;
    INT64 stack_change_out;
    INDEX iedg;
    INT64 stack_min=0;
    INT64 stack_max=0;


    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch( ifun )
        {
        case AFUN_PSEUDO_HELL:
            AFUN_FLAGS(ifun) |= FF_STK_INVALID;
            AFUN_STK_MAX(ifun) = INVALID;
            AFUN_STK_MIN(ifun) = INVALID;
            return;

        case AFUN_PSEUDO_CALL:
            AFUN_FLAGS(ifun) |= FF_STK_INVALID;
            AFUN_STK_MAX(ifun) = INVALID;
            AFUN_STK_MIN(ifun) = INVALID;
            return;

        case AFUN_PSEUDO_PAL:
            AFUN_FLAGS(ifun) &= ~FF_STK_INVALID;
            AFUN_STK_MAX(ifun) = 0;
            AFUN_STK_MIN(ifun) = 0;
            return;

        default:
            ERROR(("unexpected pseudo fun %d [%s],ifun,AFUN_NAME(ifun)\n"));
            AFUN_FLAGS(ifun) |= FF_STK_INVALID;
            AFUN_STK_MAX(ifun) = INVALID;
            AFUN_STK_MIN(ifun) = INVALID;
            return;
        }
    }


    CERROR(VALID(AFUN_MASTER(ifun)),("trying to analyze a escape function\n"));

    
    AFUN_FLAGS(ifun) |= FF_STK_INVALID;
    AFUN_STK_MAX(ifun) = INVALID;
    AFUN_STK_MIN(ifun) = INVALID;



    /* initiialize IN sets */
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_STACK_CHANGE_IN(ibbl) = BOTTOM;
        ABBL_UNMARK(ibbl);
    }

    ibbl = AFUN_BBL_FIRST(ifun);
    
    ABBL_STACK_CHANGE_IN(ibbl) = 0;
    ABBL_MARK(ibbl);
    
    /* do fixpoint computation */

    do
    {
        repeat = 0;
        FOREACH_FUN_BBL(ifun,ibbl)
        {

            if( !ABBL_IS_MARKED(ibbl) ) continue;
            ABBL_UNMARK(ibbl);

            if( ABBL_IS_LAST(ibbl) ) continue;

            /* compute new out value from the in value of a bbl */
            
            stack_change_out = ABBL_STACK_CHANGE_IN(ibbl);
            CASSERT( VALID(stack_change_out),("weird\n"));
            CASSERT((stack_change_out & 0x7) == 0,("unaligned stack in fun %d [%s]\n",ifun,AFUN_NAME(ifun)));

            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_REG_C(iins) != REG_SP ) continue;
                
                if( AINS_TYPE(iins) != IT_LDA || AINS_REG_B(iins) != REG_SP )
                {
#ifdef DEBUG_STK
                    WARNING(("unexpected stack manipulation in bbl %d fun %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
                    PrintIns(stdout,iins);
#endif
                    return;
                }
                
                if( AINS_CODE(iins) == I_lda )
                {
                    stack_change_out += AINS_DATA(iins);
                    CASSERT((stack_change_out & 0x7) == 0,("unaligned stack in fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
                    if( stack_change_out != 0 )
                    {
                        if( stack_change_out > 0 )
                        {
#ifdef DEBUG_STK
                            WARNING(("stack grows positively (%d) in fun %d [%s]\n",stack_change_out,ifun,AFUN_NAME(ifun)));
                            PrintBblSuper(stdout,ibbl);
#endif
                        }
                        
                        if( stack_min == 0 || stack_min > stack_change_out )
                        {
                            stack_min = stack_change_out;
                        }

                        if( stack_max == 0 || stack_max < stack_change_out )
                        {
                            stack_max = stack_change_out;
                        }
                    }
                }
                else
                {
                    ERROR(("weird stack updating instruction\n"));
                    PrintIns(stdout,iins);
                    return;
                }
            }

            /*
              the following code makes up for the fact that function
               calls generated by critical path escapes seemingly violate
               the stack updating
               */
               
               
            if( ABBL_TYPE(ibbl)==BT_CALL && ABBL_IS_CRIT_ESC(ibbl) )
            {
                INDEX jfun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
                CASSERT( VALID(AFUN_MASTER(jfun)),("weird\n"));
                CERROR( AFUN_HAS_STK_INVALID(jfun),("invalid stack\n"));
                
                if( !AFUN_HAS_STK_INVALID(jfun) )
                {
                    CASSERT( AFUN_STK_MIN(jfun) == AFUN_STK_MAX(jfun),("weird\n"));
#if 0
                    WARNING(("USED STACKINFO %d FOR CRIT_ESC to %d [%s] in %d [%s]\n",
                             AFUN_STK_MAX(jfun),jfun,AFUN_NAME(jfun),ifun,AFUN_NAME(ifun)));
#endif
                    stack_change_out -= AFUN_STK_MAX(jfun);
                }
                
            }
            

            CASSERT((stack_change_out & 0x7) == 0,("unaligned stack %ld in fun %d [%s]\n",stack_change_out,ifun,AFUN_NAME(ifun)));
                    
            /* propagate out value of bbl to in's of its successors */
            
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;

                succ = AEDG_TAIL(iedg);

                /* ignore escape edges need to be handled in routine that uses stack info */
                
                if( ABBL_FUN(succ) != ifun )
                {
                    continue;
                }
                
                if( ABBL_STACK_CHANGE_IN(succ) == BOTTOM )
                {
                    ABBL_STACK_CHANGE_IN(succ) = stack_change_out;
                    ABBL_MARK(succ);
                    repeat = 1;
                
                }
                else if ( ABBL_STACK_CHANGE_IN(succ) != stack_change_out )
                {

#ifdef DEBUG_STK
                    WARNING(("inconsistent stack updates %d<->%d reaching bbl: %d in fun: %d [%s]\n",
                             ABBL_STACK_CHANGE_IN(succ),stack_change_out,ibbl,ifun,AFUN_NAME(ifun)));
#endif
                    return;
                    
                }
                else
                {
                    continue;
                }
            }
        }
    }
    while(repeat);

    ibbl = AFUN_BBL_LAST(ifun);
    CASSERT( ABBL_TYPE(ibbl) == BT_EXIT,("last bbl not an exit bbl in fun: in %d [%s]\n",ifun,AFUN_NAME(ifun)));

    
    if( ABBL_STACK_CHANGE_IN(ibbl) != 0 )
    {
#ifdef DEBUG_STK
        WARNING(("inconsistent stack update %d in last bbl in fun: %d [%s]\n",ABBL_STACK_CHANGE_IN(ibbl),ifun,AFUN_NAME(ifun)));
        PrintBblSuper(stdout,ibbl);
#endif
        return;
    }

    AFUN_FLAGS(ifun) &= ~FF_STK_INVALID;
    AFUN_STK_MIN(ifun) = stack_min;
    AFUN_STK_MAX(ifun) = stack_max;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunComputeStackUsage"

static void AfunComputeStackUsage(INDEX ifun)
{
    INDEX ibbl;
    INDEX iins;
    NUMBER offset;
    INT32  stack_curr;

    CERROR(VALID(AFUN_MASTER(ifun)),("trying to analyze a escape function\n"));
        
    if( AFUN_IS_PSEUDO(ifun) )
    {
        switch( ifun )
        {
        case AFUN_PSEUDO_HELL:
            AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
            return;
        case AFUN_PSEUDO_CALL:
            AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
            return;
        case AFUN_PSEUDO_PAL:
            AFUN_FLAGS(ifun) |= FF_STK_ARG;
            AFUN_FLAGS(ifun) &= ~FF_STK_ALIAS;
            return;
        default:
            ERROR(("unexpected pseudo fun %d [%s],ifun,AFUN_NAME(ifun)\n"));
            AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
            return;
        }
    }

        
    if( AFUN_FLAGS(ifun) & FF_STK_INVALID )
    {
        AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
        return;
    }

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_STACK_CHANGE_IN(ibbl) == BOTTOM )
        {
            AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
            return;
        }
    }
    

    AFUN_FLAGS(ifun) &= ~(FF_STK_ARG|FF_STK_ALIAS);
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        stack_curr = ABBL_STACK_CHANGE_IN(ibbl);
        
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_REG_A(iins) != REG_SP && AINS_REG_B(iins) != REG_SP && AINS_REG_C(iins) != REG_SP )  continue;

            switch( AINS_TYPE(iins) )
            {
            case IT_LDA:
                
                if( AINS_REG_C(iins) == REG_SP && AINS_REG_B(iins) == REG_SP )
                {
                    CASSERT( AINS_CODE(iins) == I_lda,("lda expected\n"));
                    CASSERT( AINS_REG_B(iins) == REG_SP,("reg_bis not reg_sp\n"));
                    stack_curr += AINS_DATA(iins);
                    break;
                }

                offset = stack_curr + AINS_DATA(iins);
                        
                if( AINS_REG_B(iins) == REG_SP )
                {
#if 0
                    AFUN_FLAGS(ifun) |= FF_STK_ALIAS | FF_STK_ARG;
#else
                    AFUN_FLAGS(ifun) |= FF_STK_ALIAS;
#endif
                    
#ifdef DEBUG_STK
                    CWARNING( AINS_DATA(iins) < 0 ,("negative stack offset found in %d [%s]\n",ifun,AFUN_NAME(ifun)));
#endif
                    if( offset >= 0  )
                    {
                        AFUN_FLAGS(ifun) |= FF_STK_ARG;
#ifdef DEBUG_STK
                        CWARNING(!AFUN_HAS_STK_ARG(ifun),("weird stack arg found in %d [%s]\n",ifun,AFUN_NAME(ifun)));
                        PrintIns(stdout,iins);
#endif
                        
                    }

                    if( offset < AFUN_STK_MIN(ifun) )
                    {
#if 0
                        WARNING(("out of bounds stack access in %d [%s]\n",ifun,AFUN_NAME(ifun)));
                        
#endif
                        AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
                        return;
                    }
                }
                else
                {
#ifdef DEBUG_STK
                    WARNING(("strange stack usage in fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
                    PrintIns(stdout,iins);
#endif
                    AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
                    return;
                }
                

                break;
                                
            case IT_ILD:
            case IT_IST:
            case IT_FLD:
            case IT_FST:
            case IT_ULD:
            case IT_UST:

                offset = stack_curr + AINS_DATA(iins);
                
                if( offset < AFUN_STK_MIN(ifun) )
                {
#if 1
                    WARNING(("out of bounds stack access in %d [%s]\n",ifun,AFUN_NAME(ifun)));
                    
#endif
                    AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
                    return;
                }
                

                if( offset + AinsMemWidth(iins) > 0  )
                {
#ifdef DEBUG_STK
                    CWARNING(!AFUN_HAS_STK_ARG(ifun),("stack arg found in %d [%s] current stk: %d offset %d\n",
                                                      ifun,AFUN_NAME(ifun),stack_curr,AINS_DATA(iins)));
                    PrintIns(stdout,iins);
#endif                    
                    AFUN_FLAGS(ifun) |= FF_STK_ARG;
                }
                break;


            default:
#ifdef DEBUG_STK
                WARNING(("strange stack usage in fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
                PrintIns(stdout,iins);
#endif
                AFUN_FLAGS(ifun) |= FF_STK_ARG | FF_STK_ALIAS;
                return;
            }
                
        }
    }
    return;
}

/*=======================================================
@COM
Analyze stack size and usage.
This anlysis determines the stack behaviour of all the functions.
Upon completion

AFUN_MIN(fun) contains the maximum stackframe size used my the function
              (this is a negative number).
AFUN_MAX(fun) if the stackframe size varies this is the smallest
              stack framesize encountered.
              (this is a negative number).

For each basic block in fun

ABBL_STACK_CHANGE_IN(bbl) contains the stackpointer relative to what it
                          was at function entry.
                          Stackmerging invalidates this field.

                          
A variety of flags are set or cleared for functions:

AFUN_HAS_STK_ARG(fun)        the function accesses the stack outside  of its frame or has a stack alias
AFUN_HAS_STK_ALIAS(ifun)     the function copies the value of the stackpointer
                             plus an offset to another register (using lda)
AFUN_HAS_STK_BAD_ALIAS(ifun) the function uses the value of the stack pointer with
                             an instruction different from lda.
AFUN_HAS_STK_INVALID         the function does bad things with the stack
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "ComStackInfo"

extern void ComStackInfo(FILE *fp)    
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Stackinfo");
    INDEX ifun;
    NUMBER iv = 0;
    NUMBER ivh = 0;
    NUMBER ze = 0;
    NUMBER zeh = 0;
    NUMBER nz = 0;
    NUMBER nzh = 0;    
    NUMBER sa = 0;
    NUMBER sah = 0;
    NUMBER al = 0;
    NUMBER alh = 0;
    BOOL hot;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        if( VALID(AFUN_MASTER(ifun)) ) continue;
        AfunComputeStackChange(ifun);
        AfunComputeStackUsage(ifun);
    }

    
    FOREACH_LIVE_FUN(ifun)
    {
        if( VALID(AFUN_MASTER(ifun)) ) continue;
            
        hot = (AfunNBblsHot(ifun) > 0);

        if( verbose )
        {
            fprintf(fp,"\t[*] %d [%s] min %d max %d ",ifun,AFUN_NAME(ifun),AFUN_STK_MIN(ifun),AFUN_STK_MAX(ifun));
            if( hot ) fprintf(fp,"hot ");
            
        }
        
        if( AFUN_HAS_STK_INVALID(ifun) )
        {
            if( verbose ) fprintf(fp,"invalid ");
            iv++;
            if( hot ) ivh++;
        }

        if( AFUN_STK_MIN(ifun) == 0 )
        {
            ze++;
            if( hot ) zeh++;
        }
        else
        {    
            nz++;
            if( hot ) nzh++;
        }
        
        if( AFUN_HAS_STK_ALIAS(ifun) )
        {
            if( verbose ) fprintf(fp,"stk_alias ");
            al++;
            if( hot ) alh++;
        }

        
        if( AFUN_HAS_STK_ARG(ifun) )
        {
            if( verbose ) fprintf(fp,"stk_arg ");
            sa++;
            if( hot ) sah++;
        }

        if( verbose ) fprintf(fp,"\n");
    }

    
            
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    InfoStatus(fp,"\t[ComputeStackinfo] iv %d#%d nz %d#%d ze %d#%d  al %d#%d sa %d#%d {%d ms}\n",
            iv,ivh,nz,nzh,ze,zeh,al,alh,sa,sah,time);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunMergeStack"

static void AfunMergeStack(INDEX ifun)
{
    INDEX  ibbl;
    INDEX  iins;
    INT32  stack_curr;
    INT32  offset;
    INT32 stack_new = AFUN_STK_MIN(ifun);

    CERROR( AFUN_HAS_STK_INVALID(ifun),("cannot merge invalid stack\n"));
    CERROR( AFUN_HAS_STK_ALIAS(ifun),("cannot merge bad alias stack\n"));
    CERROR( VALID(AFUN_MASTER(ifun)),("weird\n"));
        
    CASSERT( stack_new > -(1<<15)  && stack_new < (1<<15),("bad stack size: %lx\n",stack_new));
    
    /* now merge the stacks */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        stack_curr = ABBL_STACK_CHANGE_IN(ibbl);    


#if 0
        printf("%d %d\n",ibbl,stack_curr);
#endif   
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_REG_A(iins) != REG_SP && AINS_REG_B(iins) != REG_SP && AINS_REG_C(iins) != REG_SP )  continue;

            switch( AINS_TYPE(iins) )
            {
            case IT_LDA:
                if( AINS_REG_C(iins) == REG_SP )
                {
                    CASSERT( AINS_CODE(iins) == I_lda,("lda expected\n"));
                    CASSERT( AINS_REG_B(iins) == REG_SP,("reg_bis not reg_sp\n"));
                    stack_curr += AINS_DATA(iins);
                    AinsKill(iins);
                    break;
                }


                offset = AINS_DATA(iins) ;
                if( offset <  0 )
                {
                    WARNING(("negative stack offset %d (before) in bbl %d\n",offset,ibbl));
                    PrintBblSuper(stdout,ibbl);
                }
                
                offset += (stack_curr - stack_new);
                if( offset <  0 )
                {
                    WARNING(("negative stack offset %d (after) in bbl %d\n",offset,ibbl));
                    PrintBblSuper(stdout,ibbl);
                }
                CASSERT( offset  > -(1<<15)  && offset < (1<<15),("bad offset: %lx\n",stack_new));
                AINS_DATA(iins) = offset;
                break;
                
            case IT_ILD:
            case IT_IST:
            case IT_FLD:
            case IT_FST:
            case IT_ULD:
            case IT_UST:
                offset = AINS_DATA(iins) ;
#if 0
                printf("old offset: %04lx ",offset);
#endif
                if( offset <  0 )
                {
                    WARNING(("negative stack offset %d (before) in bbl %d\n",offset,ibbl));
                    PrintBblSuper(stdout,ibbl);
                }
                
                offset += (stack_curr - stack_new);
                if( offset <  0 )
                {
                    WARNING(("negative stack offset %d (after) in bbl\n",offset,ibbl));
                    PrintBblSuper(stdout,ibbl);
                }
                CASSERT( offset  > -(1<<15)  && offset < (1<<15),("bad offset: %lx\n",stack_new));
                AINS_DATA(iins) = offset;
#if 0
                printf("new offset: %04lx\n",offset);
#endif
                if( offset > -stack_new - AinsMemWidth(iins) )
                {
                    WARNING(("access out of of stk frame offset %x stack %x in bbl %d\n",offset,-stack_new,ibbl));
                    CASSERT( AFUN_HAS_STK_ARG(ifun),("stack arg flag expected for fun %d [%s]\n",ifun,AFUN_NAME(ifun)));
                    /*
                    PrintBblSuper(stdout,ibbl);
                    */
                    
                }
                break;
            default:
                ERROR(("unexpected strange stack usage in bbl %d\n",ibbl));
                PrintIns(stdout,iins);
                break;
            }
            
        }


        /*
          the following code helps with to compensate for calls to
          functions with stack aliases or escape functions
          It destroys the ABBL_STACK_CHANGE_IN information !!!
          */
        
        if( ABBL_TYPE(ibbl) == BT_CALL )
        {
            ABBL_STACK_CHANGE_IN(ibbl) = stack_curr;
        }
        
        if( ABBL_IS_FIRST(ibbl) )
        {
            iins = AinsNew();
            AinsMakeLda(iins,REG_SP,REG_SP,stack_new);
            AINS_BBL(iins) = ibbl;
            AinsPrepend(iins);
        }

        if( ABBL_TYPE(ibbl) == BT_RET )
        {
            INDEX last = ABBL_INS_LAST(ibbl);
            iins = AinsNew();
            AinsMakeLda(iins,REG_SP,REG_SP,-stack_new);
            AINS_BBL(iins) = ibbl;
            CASSERT( AINS_CODE(last) == I_ret,("last iins is not I_ret\n"));
            AinsInsertBefore(iins,last);
        }
    }


    /*
      the following code is necessary to compensate for calls to
      functions with stack aliases or escape functions
      */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        
        if( ABBL_TYPE(ibbl) == BT_CALL )
        {
            INDEX last = ABBL_INS_LAST(ibbl);
            INDEX target = AEDG_TAIL(ABBL_EDG_CALL(ibbl));
            INDEX callee = ABBL_FUN(target);
            INDEX next = AEDG_TAIL(ABBL_EDG_LINK(ibbl));
            
            if( AFUN_HAS_STK_ARG(callee) || ABBL_IS_CRIT_ESC(ibbl) )  
            {
                
                CASSERT(AINS_CODE(last)==I_bsr || AINS_CODE(last)==I_jsr,("last instruction is not bsr/jsr\n"));

                if( AbblNpreds(next) != 2 )
                {
                    INDEX next_new = AbblNew( );
                    HURRAH(("creating a new bbl\n"));
                    ABBL_TYPE(next_new) = BT_NORM;
                    ABBL_FUN(next_new) = ifun;
                    AedgLinkFancy(AedgNewFancy(next_new,next,ET_NORM));

                    AedgKillFancy( ABBL_EDG_CALL(ibbl));
                    AedgKillFancy( ABBL_EDG_LINK(ibbl));
                                   
                    AedgLinkFancy( AedgNewFancy(ibbl,next_new,ET_FUN_LINK));
                    AedgLinkFancy( AedgNewFancy(ibbl,target,ET_FUN_CALL));

                    ABBL_STACK_CHANGE_IN(next_new) = ABBL_STACK_CHANGE_IN(next);
                    AbblAppendFancy(next_new);                    
                    next = next_new;
                }
                
                stack_curr = ABBL_STACK_CHANGE_IN(ibbl);    /* we saved this above, remember */
                
                if( stack_curr-stack_new != 0 )
                {
                    WARNING(("PRECALL %d [%s]: adding lda's %x current %x new %x in bbl %d\n",
                             callee,AFUN_NAME(callee),stack_curr-stack_new,stack_curr,stack_new,ibbl));
                    iins = AinsNew();
                    AinsMakeLda(iins,REG_SP,REG_SP,stack_curr-stack_new);
                    AINS_BBL(iins) = ibbl;
                    AinsInsertBefore(iins,last);
                    /*
                    PrintBblSuper(stdout,ibbl);
                    */
                }
                
                /* because of escape functions the next stack curr might be different */

                stack_curr = ABBL_STACK_CHANGE_IN(next);
                if( stack_curr-stack_new != 0 )
                {
                    WARNING(("POSTCALL: adding lda's in bbl %x current %x new %x in bbl %d\n",
                             stack_curr-stack_new,stack_curr,stack_new,ibbl));
                    iins = AinsNew();
                    AinsMakeLda(iins,REG_SP,REG_SP,stack_new-stack_curr);
                    AINS_BBL(iins) = next;
                    AinsPrepend(iins);
                    /*
                    PrintBblSuper(stdout,next);
                    */
                }
            }
            
        }
        
    }

}



/*=======================================================
@FUN
obsolete
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunCallsBadFunctions"

extern BOOL AfunCallsBadFunction(INDEX ifun)    
{
    INDEX ibbl;
    INDEX callee;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL ) continue;
        callee = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));

        if( ABBL_IS_CRIT_ESC(ibbl) )
        {
            WARNING(("CRIT_ESC FUN needs work call to %d [%s]\n",callee,AFUN_NAME(callee)));
        }
        
        
        if( AFUN_HAS_STK_ARG(callee) )
        {
            WARNING(("STK_ARG FUN needs work call to %d [%s]\n",callee,AFUN_NAME(callee)));
        }
        
    }
    return 0;
}

/*=======================================================
@OPT
Merge stacks.
This optimization tries to reduce the number of stack pointer
manipulations to one at the beginning of the entry bbl
and one before each return instruction.

requires  ComStackInfo
@END
=======================================================*/


#undef __PROC__
#define __PROC__ "OptStackMerge"

extern BOOL OptStackMerge(FILE *fp)    
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Stackmerge");
    BOOL   change;
    NUMBER count=0;
    INDEX  ifun;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        if( AFUN_IS_PSEUDO(ifun) ) continue; 
        if( VALID(AFUN_MASTER(ifun)) ) continue;
        if( AfunHas2Esc(ifun) ) continue;
        if( AfunHasEsc2(ifun) ) continue;
        if( AFUN_HAS_STK_INVALID(ifun) ) continue;
        if( AFUN_HAS_STK_ALIAS(ifun) ) continue;

        /* let's see whether the stack needs to be merged in this one */

        if( AFUN_STK_MAX(ifun) == AFUN_STK_MIN(ifun) ) continue;

        if( verbose )
        {
            fprintf(fp,"\t[*] stack size range from %d to %d in fun: %d [%s]\n",
                    AFUN_STK_MIN(ifun),AFUN_STK_MAX(ifun),ifun,AFUN_NAME(ifun));
        }

        
        AfunMergeStack(ifun);
        count++;
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (count>0);

    InfoStatus(fp,"\t[OptStackmerge%c] Merged stks: %d {%d ms}\n",CHANGE(change),count,time);
        
    return change;
}

/*=======================================================
@OPT
Removes obviously unnecessary stack stores

requires  ComStackInfo
@END
=======================================================*/

typedef struct
{
    NUMBER nreads;
    NUMBER nwrites;
    FLAGS  readtype;
    FLAGS writetype;
}
STACKENTRY;


#undef __PROC__
#define __PROC__ "OptKillDeadStores"

extern BOOL OptKillDeadStores(FILE *fp)    
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Killdeadstores");
    BOOL   change;
    INDEX  ifun;
    INDEX  ibbl;
    INDEX  iins;
    INT32  stack_curr;
    NUMBER i;
    INT32  offset;
    NUMBER stacksize;
    NUMBER memwidth;
    STACKENTRY *Stack;

    NUMBER count=0;
    NUMBER counth=0;
    INT64 weight=0;

    
    clock_t time = clock();
    
#if 0
    PrintFunSuper(stdout,35);
    PrintFunSuper(stdout,40);
#endif
    FOREACH_LIVE_FUN(ifun)
    {
        if( AFUN_IS_PSEUDO(ifun) ) continue; 
        if( VALID(AFUN_MASTER(ifun)) ) continue;
        if( AfunHas2Esc(ifun) ) continue;
        if( AfunHasEsc2(ifun) ) continue;
        if( AFUN_HAS_STK_INVALID(ifun) ) continue;
        if( AFUN_HAS_STK_ALIAS(ifun) ) continue;
        if( AFUN_HAS_STK_ARG(ifun) ) continue;
        if( AFUN_STK_MIN(ifun) == 0 ) continue;
        if( AfunCallsStackArgFuns(ifun) ) continue;

        if( verbose )
        {
            InfoExtra(fp,"\t[*] stack size range from %d to %d in fun: %d [%s]\n",
                    AFUN_STK_MIN(ifun),AFUN_STK_MAX(ifun),ifun,AFUN_NAME(ifun));
        }

        stacksize = -AFUN_STK_MIN(ifun);
        
        CASSERT( stacksize > -(1<<15)  && stacksize < (1<<15),("bad stack size: %lx\n",stacksize));

        Stack = calloc(sizeof(STACKENTRY),stacksize);     

        FOREACH_FUN_BBL(ifun,ibbl)
        {
            stack_curr = ABBL_STACK_CHANGE_IN(ibbl);    

            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_REG_A(iins) != REG_SP && AINS_REG_B(iins) != REG_SP && AINS_REG_C(iins) != REG_SP )  continue;
            
                switch( AINS_TYPE(iins) )
                {
                case IT_LDA:
                    if( AINS_REG_C(iins) == REG_SP )
                    {
                        CASSERT( AINS_CODE(iins) == I_lda,("lda expected\n"));
                        CASSERT( AINS_REG_B(iins) == REG_SP,("reg_b is  not reg_sp\n"));
                        stack_curr += AINS_DATA(iins);
                    }
                    break;

                case IT_ILD:
                case IT_FLD:
                    memwidth = AinsMemWidth(iins);
                    offset = AINS_DATA(iins) + stack_curr;
                    if( offset+memwidth >  0 ||  offset < -stacksize )
                    {
                        WARNING(("illegal stack offset %d in bbl %d\n",offset,ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
                        fflush(stdout);
                    }

                    for(i=0;i<memwidth;i++)
                        Stack[-(offset+i)].nreads++;
                    
                    break;
                    
                case IT_FST:
                case IT_IST:
                    memwidth = AinsMemWidth(iins);
                    offset = AINS_DATA(iins) + stack_curr;
                    if( offset+memwidth >  0 ||  offset < -stacksize )
                    {
                        WARNING(("illegal stack offset %d in bbl %d\n",offset,ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
                        fflush(stdout);
                    }
                    for(i=0;i<memwidth;i++)
                        Stack[-(offset+i)].nwrites++;
                    break;
                    
                default:
                    ERROR(("unexpected strange stack usage in bbl %d\n",ibbl));
                    PrintIns(stdout,iins);
                    break;
                }
                
            }
        }

        /* eliminate unecessary store */
        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            stack_curr = ABBL_STACK_CHANGE_IN(ibbl);    
        
            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_REG_A(iins) != REG_SP && AINS_REG_B(iins) != REG_SP && AINS_REG_C(iins) != REG_SP )  continue;
            
                switch( AINS_TYPE(iins) )
                {
                case IT_LDA:
                    if( AINS_REG_C(iins) == REG_SP )
                    {
                        CASSERT( AINS_CODE(iins) == I_lda,("lda expected\n"));
                        CASSERT( AINS_REG_B(iins) == REG_SP,("reg_b is  not reg_sp\n"));
                        stack_curr += AINS_DATA(iins);
                    }
                    break;
                    
                case IT_ILD:
                case IT_FLD:
                    memwidth = AinsMemWidth(iins);
                    offset = AINS_DATA(iins) + stack_curr;
                    if( offset+memwidth >  0 ||  offset < -stacksize )
                    {
                        WARNING(("illegal stack offset %d in bbl %d\n",offset,ibbl));
                        PrintBblSuper(stdout,ibbl);
                    }
                    
                    for(i=0;i<memwidth;i++)
                    {
                        if( Stack[-(offset+i)].nwrites > 0 )
                            break;
                    }
                    
                    if( i == memwidth )
                    {
#if 1
                        printf("QQQ read without write\n");
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
#endif
                    }
                
                    break;
                    
                case IT_FST:
                case IT_IST:
                    memwidth = AinsMemWidth(iins);
                    offset = AINS_DATA(iins) + stack_curr;
                    if( offset+memwidth >  0 ||  offset < -stacksize )
                    {
                        WARNING(("illegal stack offset %d in bbl %d\n",offset,ibbl));
                        PrintBblSuper(stdout,ibbl);
                        fflush(stdout);
                    }

                    for(i=0;i<memwidth;i++)
                    {
                        if( Stack[-(offset+i)].nreads > 0 )
                        break;
                    }

                    if( i == memwidth )
                    {
#if 0
                        printf("QQQ write without read - killed %ld\n",ABBL_COUNT(ibbl));
                        PrintIns(stdout,iins);
                        PrintBblSuper(stdout,ibbl);
#endif
                        AinsKill(iins);
                        count++;
                        if( ABBL_IS_HOT(ibbl) )
                            counth++;
                        weight += ABBL_COUNT(ibbl);
                    }
                    
                break;
                
                default:
                    ERROR(("unexpected strange stack usage in bbl %d\n",ibbl));
                    PrintIns(stdout,iins);
                    break;
                }
                
            }
        }

        free(Stack);
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (count>0);

    InfoStatus(fp,"\t[OptKilldeadstores%c] Killed Stores: %d#%d  weight #%ld {%d ms}\n",CHANGE(change),count,counth,weight,time);

    return change;
}

/*=======================================================
=======================================================*/




