/*=======================================================
  Author:   
  Version:  $Id$
  Comment:  Code for loading profiling info
=======================================================*/



#include "global.h"
#include "functions.h"
#include "flags.h"

#ifdef SYS_COFF
#include <stdio.h>
#include <cmplrs/prof_header.h>
#include <cmplrs/pdf.h>
#endif

#define FILE_PDF 0x0a213e4644503c21L
#define FILE_ADR 0x0f0e0010L
#define FILE_CNT 0x0f0e0a11L
#define FILE_CNT2 0x0f0e0a33L


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeReadProfilingInformation"

#define USE_NEW_ATOM_PIXIE 1


typedef struct {
    ADDRESS addr;
    ADDRESS count;
} indir2;

typedef struct {
    INT64 iins;
    INT64 totalcount;
    indir2 values[CONFIG_NINDIR];
} Indir2;



Indir *aIndir;
NUMBER aIndirFuns;

extern void ComputeReadProfilingInformation(FILE *fp)
{
    FILE *fp_adr;
    char fn_adr[256];
    
    CASSERT( CONFIG_USE_PROFILE,("weird\n"));
    

    /*
      check to see if we have an alto counts file, and
      if so, use that instead of the pixie data
      */
    sprintf(fn_adr,"%s.AltoCounts",CONFIG_INFILE);
    fp_adr = fopen(fn_adr,"r");
    if (fp_adr != 0)
    {
        fclose(fp_adr);
	ComputeReadAltoProfilingInformation(fp);
    }
    else
    {
        ComputeReadPixieProfilingInformation(fp);
    }
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComputeReadPixieProfilingInformation"

extern void ComputeReadPixieProfilingInformation(FILE *fp)
{
    INDEX iedg;
    INDEX ibbl;
    FILE *fp_adr;
    FILE *fp_cnt;
    char fn_adr[256];
    char fn_cnt[256];
    int  adr;
    INT64 cnt;
    NUMBER rcount = 0;
    NUMBER gcount = 0;
    
    CASSERT( CONFIG_USE_PROFILE,("weird\n"));
    

    /* otherwise, use the pixie data */

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    sprintf(fn_adr,"%s.Addrs",CONFIG_INFILE);

    fp_adr = fopen(fn_adr,"r");
    CFATAL(fp_adr==0,("Cannot open adr file: %s\n",fn_adr));
    fread(&cnt,sizeof(cnt),1,fp_adr);

    if( (cnt & 0xffffffff) == FILE_ADR )
    {
        FATAL(("Your are using and old pixie version\n"));
    }
    else if( cnt == FILE_PDF  )
    {
#if USE_NEW_ATOM_PIXIE
        fseek(fp_adr,0x100,SEEK_SET);
        /* read string length */
        fread(&cnt,sizeof(cnt),1,fp_adr);
        /* round up to multiple of 4*/
        cnt += 3;
        cnt &= ~3;

        /* skip next int32 */
        cnt += 4;
        
        /* round up to multiple of 8 */
        cnt += 7;
        cnt &= ~7;
        
        fseek(fp_adr,cnt,SEEK_CUR);
        fread(&cnt,sizeof(cnt),1,fp_adr);
        /*printf("adr (%ld),data offset: %lx\n",cnt,ftell(fp_adr));*/
#else
        FATAL(("You are using the new pixie\n"));
#endif
    }
    else
    {
        FATAL(("bad count file\n"));
    }   

    /*
      open count file and advance file pointer
      to beginning of count data
      */

    OptRemoveUnreachableCode(stdout);
    
    sprintf(fn_cnt,"%s.Counts",CONFIG_INFILE);
    
    fp_cnt = fopen(fn_cnt,"r");
    CFATAL(fp_cnt==0,("Cannot open cnt file: %s\n",fn_cnt));
    fread(&cnt,sizeof(cnt),1,fp_cnt);
    
    if( (cnt & 0xffffffff) == FILE_CNT )
    {
        FATAL(("Your are using and old pixie version\n"));
    }
    else if( (cnt & 0xffffffff) == FILE_CNT2 )
    {
        FATAL(("Your are using and old pixie version\n"));
    }
    else if( cnt == FILE_PDF  )
    {
#ifdef USE_NEW_ATOM_PIXIE
        fseek(fp_cnt,0x148,SEEK_SET);
        fread(&cnt,sizeof(cnt),1,fp_cnt);
        cnt += 3;
        cnt &= ~3;

        cnt += 4;
        cnt += 7;
        cnt &= ~7;

        fseek(fp_cnt,cnt,SEEK_CUR);
        
        fread(&cnt,sizeof(cnt),1,fp_cnt);

        /*        printf("cnt (%ld) data offset: %lx\n",cnt,ftell(fp_cnt));*/
#else
        FATAL(("You are using the new pixie\n"));
#endif
        
    }
    else
    {
        FATAL(("bad count file\n"));
    }
    
    
    for(;;)
    {
        if( 1 != fread(&cnt,sizeof(cnt),1,fp_cnt) )
            break;
        fread(&adr,sizeof(adr),1,fp_adr);


        rcount++;

        ibbl = AINS_BBL(adr);
#if 0
        printf("ins: %d  (%lx) bbl: %d cnt: %ld\n",adr,Ins2Addr(adr),ibbl,cnt);
#endif
        
               

#if 0
        if( AINS_FLAGS(adr) & FI_BBL_BEG )
        {
            ABBL_COUNT(ibbl) = cnt;
        }



        printf("iins: %d ibbl: %d \n",adr,ibbl);
        if( ibbl == 5441 )
        {
            printf("*****\n");
            
        }
#endif
        
#if 1
        if( VALID(ABBL_COUNT(ibbl)) )
        {
            if( !ABBL_IS_FIRST(ibbl) && !ABBL_IS_KILLED(ibbl) )
            {
                WARNING(("duplicate ins %d bbl %d  changing counts: %ld to %ld\n",
                         adr,ibbl,ABBL_COUNT(ibbl),cnt));
#if 0
                PrintBblSuper(stdout,ibbl);
#endif
            }
            
        }
#endif

        ABBL_COUNT(ibbl) = cnt;
    }
    
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
        if( VALID(ABBL_COUNT(ibbl)) ) continue;
        if( ABBL_TYPE(ibbl)==BT_EXIT ) continue;

        if( ABBL_TYPE(ibbl-1) == BT_PAL )
        {
            ABBL_COUNT(ibbl) = ABBL_COUNT(ibbl-1);
            continue;
        }
        
#if 1
        WARNING(("guessing count %ld for bbl %d in fun %d [%s]\n",ABBL_COUNT(ibbl),ibbl,ABBL_FUN(ibbl),AFUN_NAME(ABBL_FUN(ibbl))));
#endif
        
        gcount++;
        
        ABBL_COUNT(ibbl) = 0; 
    }

#if 1
    FOREACH_LIVE_BBL(ibbl)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
        if( ABBL_TYPE(ibbl)==BT_EXIT ) continue;
                    
        if( AbblNpreds(ibbl) == 1)
        {
            INDEX hd = AEDG_HEAD( ABBL_PRED(ibbl) );

            if( AFUN_IS_PSEUDO(ABBL_FUN(hd)) )
                continue;
            
            if( ABBL_COUNT(ibbl) > ABBL_COUNT(hd) )
            {
                WARNING(("inconsistency in bbl %d->%d  count: %ld -> %ld\n",
                       hd,ibbl,ABBL_COUNT(hd),ABBL_COUNT(ibbl)));
#if 0
                PrintBblSuper(stdout,hd);
                PrintBblSuper(stdout,ibbl);
#endif                
            }
            
        }
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            INT64 mincount = 0;
        
            switch( AEDG_TYPE(iedg) )
            {
            case ET_NORM:
            case ET_FUN_CALL:
            case ET_FUN_LINK:
                mincount += ABBL_COUNT( AEDG_HEAD(iedg) );
                break;
            }

            if( ABBL_COUNT(ibbl) < mincount )
            {
                WARNING(("inconsistency in bbl %d  is: %ld  min: %ld\n",ibbl,ABBL_COUNT(ibbl),mincount));
#if 0
                PrintBblSuper(stdout,ibbl);
#endif                
            }
        }
    }
#endif
    
    InfoStatus(fp,"\t[ComputeProfiling] Read %d Estimated: %d\n",rcount,gcount);
}
/*=======================================================
=======================================================*/


static int cmp(const void *a, const void *b)
{
    indir *v1 = (indir *)a;
    indir *v2 = (indir *)b;

    return (v2->count - v1->count);
}


extern void ComputeReadAltoProfilingInformation(FILE *fp)
{
    INDEX iedg;
    FILE *fp_adr;
    char fn_adr[256];
    INT64 cnt;
    INT64 bcount = 0;
    INT64 ecount = 0;
    INT64 fcount = 0;
    INT64 mapcount = 0;
    INT64 totalcount;
    int i;
    int j;
    int k;
    INT64 counter;
    INT64 initial;
    INT64 post;
    
    CASSERT( CONFIG_USE_PROFILE,("weird\n"));

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    sprintf(fn_adr,"%s.AltoCounts",CONFIG_INFILE);
    
    fp_adr = fopen(fn_adr,"r");
    CFATAL(fp_adr==0,("Cannot open adr file: %s\n",fn_adr));
    fread(&cnt,sizeof(cnt),1,fp_adr);


    if ( cnt == ALTO_MAGIC )
    {
#if 0
	printf("using alto profiling, checking counters\n");
#endif
	fread(&bcount,sizeof(bcount),1,fp_adr);
#if 0
	printf("number of bbl counters is %ld\n", bcount);
#endif
	fread(&ecount,sizeof(ecount),1,fp_adr);
#if 0
	printf("number of edg counters is %ld\n", ecount);
#endif
	fread(&fcount,sizeof(fcount),1,fp_adr);
#if 0
	printf("number of fun counters is %ld\n", fcount);
#endif
	fread(&mapcount,sizeof(mapcount),1,fp_adr);
#if 0
	printf("number of iins map is %ld\n", mapcount);
	
	printf("number of bbls is %d\n", CONFIG_UBBL);
#endif	
	if (bcount == 0)
        {
	    ComputeReadPixieProfilingInformation(fp);
	}
	else {
	    for (i=0; i < bcount; i++) {
		fread(&counter, sizeof(counter), 1, fp_adr);
		/* don't change exit blocks */
		if (ABBL_IS_LAST(i)) continue;
		ABBL(i).Count = counter;
	    }
	}
	
	for (i=0; i < ecount; i++)
        {
	    fread(&iedg, sizeof(counter), 1, fp_adr);
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    AEDG_COUNT(iedg) = counter;
	}

	aIndir = malloc(sizeof(Indir) * fcount);
	aIndirFuns = fcount;

	for (i=0; i < fcount; i++) {
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    aIndir[i].iins = Addr2Ins(counter);
#if 0
	    printf("reading in counts of instruction %d\n", aIndir[i].iins);
	    printf("address of instruction is %lx\n", counter);
#endif
	    fread(&counter, sizeof(counter), 1, fp_adr);
#if 0
	    printf("current counter %ld\n", counter);
#endif
	    totalcount = 0;
	    for (j=0; j < CONFIG_NINDIR; j++)
            {
		fread(&counter, sizeof(counter), 1, fp_adr);
		aIndir[i].values[j].addr = counter;
		fread(&counter, sizeof(counter), 1, fp_adr);
		aIndir[i].values[j].count = counter;
#if 0
		printf("address %lx has count of %ld\n", aIndir[i].values[j].addr, aIndir[i].values[j].count);
#endif
		totalcount += counter;
	    }
	    aIndir[i].totalcount = totalcount;
	    qsort(aIndir[i].values, CONFIG_NINDIR, sizeof(indir), &cmp);
#if 0
	    printf("indir index is %d\n", i);
	    for (j=0; j < CONFIG_NINDIR; j++) {
		printf("address %lx has count of %ld\n", aIndir[i].values[j].addr, aIndir[i].values[j].count);
	    }
#endif
	}

	/* the really clever thing to do here would be modify the address
	   in the aIndir structure based on the way this map goes */
#if 0
	printf("reading in instruction map\n");
#endif
	for (i=0; i < mapcount; i++) {
	    fread(&initial, sizeof(initial), 1, fp_adr);
	    fread(&post, sizeof(post), 1, fp_adr);
	    
	    for (j = 0; j < fcount; j++) {
		for (k = 0; k < CONFIG_NINDIR; k++) {
		    if (aIndir[j].values[k].addr == post) {
			aIndir[j].values[k].addr = initial;
		    }
		}
	    }
	    
	}

/* 	ComVerifyCounts(fp); */
	
    }
}


extern void ComVerifyCounts(FILE *fp) 
{
    INDEX ibbl;
    NUMBER tcount, fcount;

    FOREACH_BBL(ibbl) {
	if (ABBL_TYPE(ibbl) == BT_BRANCH) {
	    /* check to see if true and false edges add up */
	    tcount = AEDG_COUNT(ABBL_EDG_TRUE(ibbl));
	    fcount = AEDG_COUNT(ABBL_EDG_FALSE(ibbl));
	    
	    CERROR(((tcount + fcount) != ABBL_COUNT(ibbl)), ("ERROR true and false counts don't match bbl in bbl %d count!!!\n", ibbl));
	}
    }
}


/*=======================================================
=======================================================*/

extern void ComReadValueProfilingInformation(FILE *fp)
{
    FILE *fp_adr;
    char fn_adr[256];
    INT64 cnt;
    INT64 bcount = 0;
    INT64 ecount = 0;
    INT64 fcount = 0;
    INT64 mapcount = 0;
    INT64 totalcount;
    int i;
    int j;
    int k;
    INT64 counter;
    INT64 initial;
    INT64 post;
    Indir2 *Values;
    NUMBER ValuesFuns;
    

#if 1 
    return;
#endif

    CASSERT( CONFIG_USE_PROFILE,("weird\n"));

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    sprintf(fn_adr,"%s.ValueCounts",CONFIG_INFILE);
    
    fp_adr = fopen(fn_adr,"r");
    CERROR(fp_adr==0,("Cannot open adr file: %s\n",fn_adr));

    if (fp_adr == 0) return; 

    fread(&cnt,sizeof(cnt),1,fp_adr);


    if ( cnt == ALTO_MAGIC )
    {
	printf("using alto profiling, checking counters\n");
	fread(&bcount,sizeof(bcount),1,fp_adr);
	printf("number of bbl counters is %ld\n", bcount);
	fread(&ecount,sizeof(ecount),1,fp_adr);
	printf("number of edg counters is %ld\n", ecount);
	fread(&fcount,sizeof(fcount),1,fp_adr);
	printf("number of fun counters is %ld\n", fcount);
	fread(&mapcount,sizeof(mapcount),1,fp_adr);
	printf("number of iins map is %ld\n", mapcount);
	
	printf("number of bbls is %d\n", CONFIG_UBBL);
	
	if (bcount != 0) {
	    ERROR(("bbl count is not zero in value profile\n"));
	}

	if (ecount != 0) {
	    ERROR(("Edg count is not zero in value profile\n"));
	}

	Values = malloc(sizeof(Indir) * fcount);
	ValuesFuns = fcount;

	for (i=0; i < fcount; i++) {
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    /* bottom half of the word is the instruction, top is register */
	    Values[i].iins = counter;
	    fread(&counter, sizeof(counter), 1, fp_adr);
#if 0
	    printf("current counter %ld\n", counter);
#endif
	    totalcount = 0;
	    for (j=0; j < CONFIG_NINDIR; j++) {
		fread(&counter, sizeof(counter), 1, fp_adr);
		Values[i].values[j].addr = counter;
		fread(&counter, sizeof(counter), 1, fp_adr);
		Values[i].values[j].count = counter;
#if 0
		printf("address %lx has count of %ld\n", Values[i].values[j].addr, Values[i].values[j].count);
#endif
		totalcount += counter;
	    }
	    Values[i].totalcount = totalcount;
	    qsort(Values[i].values, CONFIG_NINDIR - 1, sizeof(indir), &cmp);
#if 0
	    printf("indir index is %d\n", i);
	    for (j=0; j < CONFIG_NINDIR; j++) {
		printf("address %16lx has count of %20ld\n", Values[i].values[j].addr, Values[i].values[j].count);
	    }
#endif

	    

	}

#if 1
	for (i=0; i < fcount; i++) {
	    INDEX ifun = ABBL_FUN(AINS_BBL(Values[i].iins & 0xffffffff));
	    if ( (i == (fcount - 1)) || (ifun != ABBL_FUN(AINS_BBL(Values[i+1].iins & 0xffffffff)))) {
		PrintFunSuper(stdout, ABBL_FUN(AINS_BBL(Values[i].iins)));
		printf("MagicString %s%d\n", CONFIG_INFILE, i);
	    }
	    printf("register profiled %ld\n", (Values[i].iins >> 32));
	    Values[i].iins = Values[i].iins & 0xffffffff;
	    printf("reading in counts of instruction %ld, has count of %ld\n", Values[i].iins, ABBL_COUNT(AINS_BBL(Values[i].iins)));
	    for (j=0; j < CONFIG_NINDIR; j++) {
		printf("address %16lx has count of %20ld\n", Values[i].values[j].addr, Values[i].values[j].count);
	    }
	    
	    ifun = ABBL_FUN(AINS_BBL(Values[i].iins));
	    if ( (i == (fcount - 1)) || (ifun != ABBL_FUN(AINS_BBL(Values[i+1].iins & 0xffffffff)))) {
		PrintFunSuper(stdout, ABBL_FUN(AINS_BBL(Values[i].iins)));
		printf("MagicString %s%d\n", CONFIG_INFILE, i);
	    }
	}
	
#endif

	/* the really clever thing to do here would be modify the address
	   in the Values structure based on the way this map goes */
	printf("reading in instruction map\n");
	for (i=0; i < mapcount; i++) {
	    fread(&initial, sizeof(initial), 1, fp_adr);
	    fread(&post, sizeof(post), 1, fp_adr);
	    
	    for (j = 0; j < fcount; j++) {
		for (k = 0; k < CONFIG_NINDIR; k++) {
		    if (Values[j].values[k].addr == post) {
			Values[j].values[k].addr = initial;
		    }
		}
	    }
	    
	}

  	exit(-1);  
    }
}

