
#ifndef REGS_H
#define REGS_H

#define REG_V0   0x00 /*  0 */
#define REG_T0   0x01 /*  1 */ 
#define REG_T1   0x02 /*  2 */
#define REG_T2   0x03 /*  3 */
#define REG_T3   0x04 /*  4 */
#define REG_T4   0x05 /*  5 */
#define REG_T5   0x06 /*  6 */
#define REG_T6   0x07 /*  7 */
#define REG_T7   0x08 /*  8 */

#define REG_S0   0x09 /*  9 */
#define REG_S1   0x0a /* 10 */
#define REG_S2   0x0b /* 11 */
#define REG_S3   0x0c /* 12 */
#define REG_S4   0x0d /* 13 */
#define REG_S5   0x0e /* 14 */
#define REG_S6   0x0f /* 15 */
#define REG_FP   0x0f /* 15 */     /* fp & s6 are the same */

#define REG_A0   0x10 /* 16 */
#define REG_A1   0x11 /* 17 */
#define REG_A2   0x12 /* 18 */
#define REG_A3   0x13 /* 19 */
#define REG_A4   0x14 /* 20 */
#define REG_A5   0x15 /* 21 */
#define REG_T8   0x16 /* 22 */
#define REG_T9   0x17 /* 23 */
#define REG_T10  0x18 /* 24 */ 
#define REG_T11  0x19 /* 25 */
#define REG_RA   0x1a /* 26 */
#define REG_PV   0x1b /* 27 */     /* pv and t5 are the same */
#define REG_T12  0x1b /* 27 */
#define REG_AT   0x1c /* 28 */
#define REG_GP   0x1d /* 29 */
#define REG_SP   0x1e /* 30 */
#define REG_ZERO 0x1f /* 31 */

#define REG_FLOAT 0x20

#define REG_F0    0x20 /* 32 */
#define REG_F1    0x21 /* 33 */
#define REG_F2    0x22 /* 34 */
#define REG_F3    0x23 /* 35 */
#define REG_F4    0x24 /* 36 */
#define REG_F5    0x25 /* 37 */
#define REG_F6    0x26 /* 38 */
#define REG_F7    0x27 /* 39 */
#define REG_F8    0x28 /* 40 */
#define REG_F9    0x29 /* 41 */
#define REG_F10   0x2a /* 42 */
#define REG_F11   0x2b /* 43 */
#define REG_F12   0x2c /* 44 */
#define REG_F13   0x2d /* 45 */
#define REG_F14   0x2e /* 46 */
#define REG_F15   0x2f /* 47 */
#define REG_F16   0x30 /* 48 */
#define REG_F17   0x31 /* 49 */
#define REG_F18   0x32 /* 50 */
#define REG_F19   0x33 /* 51 */

#define REG_FZERO 0x3f /* 63 */

#define REG_LIT 0x40  /* 64 */

#define REG2BIT(a)   (((a)==REG_LIT) ? 0UL : (1UL <<(a))) 

#define B_V0   (1UL<<0x00) /*  0 */
#define B_T0   (1UL<<0x01) /*  1 */ 
#define B_T1   (1UL<<0x02) /*  2 */
#define B_T2   (1UL<<0x03) /*  3 */
#define B_T3   (1UL<<0x04) /*  4 */
#define B_T4   (1UL<<0x05) /*  5 */
#define B_T5   (1UL<<0x06) /*  6 */
#define B_T6   (1UL<<0x07) /*  7 */
#define B_T7   (1UL<<0x08) /*  8 */

#define B_S0   (1UL<<0x09) /*  9 */
#define B_S1   (1UL<<0x0a) /* 10 */
#define B_S2   (1UL<<0x0b) /* 11 */
#define B_S3   (1UL<<0x0c) /* 12 */
#define B_S4   (1UL<<0x0d) /* 13 */
#define B_S5   (1UL<<0x0e) /* 14 */
#define B_S6   (1UL<<0x0f) /* 15 */
#define B_FP   (1UL<<0x0f) /* 15 */     /* fp & s6 are the same */

#define B_A0   (1UL<<0x10) /* 16 */
#define B_A1   (1UL<<0x11) /* 17 */
#define B_A2   (1UL<<0x12) /* 18 */
#define B_A3   (1UL<<0x13) /* 19 */
#define B_A4   (1UL<<0x14) /* 20 */
#define B_A5   (1UL<<0x15) /* 21 */
#define B_T8   (1UL<<0x16) /* 22 */
#define B_T9   (1UL<<0x17) /* 23 */
#define B_T10  (1UL<<0x18) /* 24 */ 
#define B_T11  (1UL<<0x19) /* 25 */
#define B_RA   (1UL<<0x1a) /* 26 */
#define B_PV   (1UL<<0x1b) /* 27 */     /* pv and t5 are the same */
#define B_T12  (1UL<<0x1b) /* 27 */
#define B_AT   (1UL<<0x1c) /* 28 */
#define B_GP   (1UL<<0x1d) /* 29 */
#define B_SP   (1UL<<0x1e) /* 30 */
#define B_ZERO (1UL<<0x1f) /* 31 */

#define B_FZERO (1UL<<0x3f) /* 63 */

#define B_F0    (1UL<<0x20) /* 32 */
#define B_F1    (1UL<<0x21) /* 33 */
#define B_F2    (1UL<<0x22) /* 34 */
#define B_F3    (1UL<<0x23) /* 35 */
#define B_F4    (1UL<<0x24) /* 36 */
#define B_F5    (1UL<<0x25) /* 37 */
#define B_F6    (1UL<<0x26) /* 38 */
#define B_F7    (1UL<<0x27) /* 39 */
#define B_F8    (1UL<<0x28) /* 40 */
#define B_F9    (1UL<<0x29) /* 41 */

#define B_F16    (1UL<<0x30) /* 48 */
#define B_F17    (1UL<<0x31) /* 49 */
#define B_F18    (1UL<<0x32) /* 50 */
#define B_F19    (1UL<<0x33) /* 51 */
#define B_F20    (1UL<<0x34) /* 52 */
#define B_F21    (1UL<<0x35) /* 53 */

#define SAVE_REGS (B_S0|B_S1|B_S2|B_S3|B_S4|B_S5|B_S6)
#define REGS_IARG  (B_A0|B_A1|B_A2|B_A3|B_A4|B_A5)
#define REGS_IRET  (B_V0)

#define FSAVE_REGS (B_F2|B_F3|B_F4|B_F5|B_F6|B_F7|B_F8|B_F9)
#define REGS_FARG  (B_F16|B_F17|B_F18|B_F19|B_F20|B_F21)
#define REGS_FRET  (B_F0|B_F1)

#define REGS_INT                  0x00000000ffffffffL
#define REGS_INT_BUT_ZERO         0x000000007fffffffL
#define REGS_FLT                  0xffffffff00000000L
#define REGS_FLT_BUT_ZERO         0x7fffffff00000000L

#define REGS_NONE                 0x0000000000000000L
#define REGS_ALL                  0xffffffffffffffffL
#define REGS_ALL_BUT_ZERO         0x7fffffff7fffffffL
#define REGS_PROC_INT_PRESERVED   (B_S0|B_S1|B_S2|B_S3|B_S4|B_S5|B_S6|B_RA|B_SP|B_GP)
#define REGS_PROC_FLT_PRESERVED   (B_F2|B_F3|B_F4|B_F5|B_F6|B_F6|B_F7|B_F8|B_F9)
#define REGS_PROC_INT_MAY_CHANGE  (REGS_INT & ~REGS_PROC_INT_PRESERVED)
#define REGS_PROC_FLT_MAY_CHANGE  (REGS_FLT & ~REGS_PROC_FLT_PRESERVED)
#define REGS_PAL_MAY_USE          (B_V0 | B_A0 | B_A1 | B_A2 | B_A3 | B_A4 | B_A5)
#define REGS_PAL_MUST_DEF         (B_V0 | B_A3)

#define REGS_CALL_MAY_USE         (REGS_IRET | REGS_IARG | REGS_FARG | B_PV)
#define REGS_CALL_MUST_DEF        (REGS_IRET | REGS_FRET )

#define REG_IS_INT(r)  (0<=r && r <= 31)
#define REG_IS_FLT(r)  (32<=r && r <= 63)

#define REGSET_FOREACH_REG(regset,reg)   \
    for( reg = 0; reg <= 63; reg++)       \
      if( ! REGSET_IS_MEMBER(reg,regset) ) \
         continue; else

#define REGSET_SINGLETON(reg) REG2BIT(reg)

#define REGSET_UNION(rs1,rs2) ((rs1) | (rs2))
#define REGSET_INTERSECT(rs1,rs2) ((rs1) & (rs2))
#define REGSET_COMPL(rs) (~(rs))
#define REGSET_DIFF(rs1,rs2) REGSET_INTERSECT((rs1),REGSET_COMPL(rs2))
#define REGSET_ADD(reg,rs) REGSET_UNION(REGSET_SINGLETON(reg),(rs))

#define REGSET_IS_EMPTY(rs) ((rs) == 0UL)
#define REGSET_IS_SUBSET(subrs,rs) ((subrs) == ((subrs) & (rs)))
#define REGSET_IS_MEMBER(member,rs) REGSET_IS_SUBSET(REGSET_SINGLETON(member),(rs))
     
#endif /* REGS_H */
