/*
**	This file contains macros, type and structure
**	defintions that should eventually go into global.h
*/

/* ---------------------------------------------------- */
/* The pipelines of 21164.  PL_NP = "no pipline" */ 

#define PL_NP           0
#define PL_E0           1
#define PL_E1           2
#define PL_FA           4
#define PL_FM           8
#define PL_E0_E1        (PL_E0 | PL_E1)
#define PL_FA_FM        (PL_FA | PL_FM)
#define PL_E0_E1_FA_FM  (PL_E0_E1 | PL_FA_FM)

#define NO_OF_PIPELINES 4

enum functional_units { FU_IMULTIPLIER, FU_FDIVIDER, NO_OF_FUNC_UNITS };

typedef enum functional_units  FUNC_UNIT;
typedef struct
{
   STRING    name;
   PIPELINES pipelines;
   LATENCY   latency1;
   LATENCY   latency2;
   LATENCY   muladd;
   BOOL      special;
} SCH_BASIC_TABLE_ENTRY;

#define NoLat INVALID

/* ----------------------------------------------------------------- */

extern SCH_BASIC_TABLE_ENTRY SchBasicTable[];

extern LATENCY ProdConsLatency [NO_OF_SCHED_CLASSES] [NO_OF_SCHED_CLASSES];
extern LATENCY ProdProdLatency [NO_OF_SCHED_CLASSES] [NO_OF_SCHED_CLASSES];
extern LATENCY FuncUnitUse     [NO_OF_FUNC_UNITS   ] [NO_OF_SCHED_CLASSES];

/* ----------------------------------------------------------------- */


#define SCHED_MAX_INSTRS 10000

#define MAX(a, b) ((a) >= (b) ? (a) : (b))

/*
**	Non zero arc values (latencies) are scaled by this amount,
**	since ordinary ref/def constraints that are zero valued
**	get converted to one.
*/

#define SCALE_FACTOR 4

#define USES FALSE
#define SETS TRUE

/* ----------------------------------------------------------------- */

#define SCHCL_NAME( SCHCL )        SchBasicTable[SCHCL].name
#define SCHCL_PIPELINES( SCHCL )   SchBasicTable[SCHCL].pipelines
#define SCHCL_LATENCY1( SCHCL )    SchBasicTable[SCHCL].latency1
#define SCHCL_LATENCY2( SCHCL )    SchBasicTable[SCHCL].latency2
#define SCHCL_MULADD( SCHCL )      SchBasicTable[SCHCL].muladd
#define SCHCL_IS_SPECIAL( SCHCL )  SchBasicTable[SCHCL].special

#define  LAST_SCHED_CLASS NO_OF_SCHED_CLASSES-1
#define FIRST_SCHED_CLASS 0

#define FOREACH_SCHED_CLASS(SCHCL) \
   for( SCHCL = 0 ; SCHCL < NO_OF_SCHED_CLASSES ; SCHCL++ )

#define FOREACH_FUNC_UNIT(FU) \
   for( FU = 0 ; FU < NO_OF_FUNC_UNITS ; FU++ )

#define  LAST_FUNC_UNIT NO_OF_FUNC_UNITS-1
#define FIRST_FUNC_UNIT 0

/* ----------------------------------------------------------------- */
     extern SCHED_CLASS AINS_SCHEDCLASS(INDEX i);
     
#define AINS_SCHEDCLASSNAME(i)  SCHCL_NAME(AINS_SCHEDCLASS(i))

#define SCHIN_INST(i)    SchedInstr[i]
#define SCHIN_OPCODE(i)  AINS_CODE(SCHIN_INST(i))
#define SCHIN_BASEREG(i) AINS_REG_B(SCHIN_INST(i))
#define SCHIN_OFFSET(i)  AINS_DATA(SCHIN_INST(i))
#define SCHIN_CLASS(i)   AINS_SCHEDCLASS(SCHIN_INST(i))
#define SCHIN_PIPELINES(i) SCHCL_PIPELINES(SCHIN_CLASS(i))
#define SCHIN_IS_PROCESSED(i) (i == FAKE_INSTR_BEFORE_FIRST || \
	                       instr_processed[i] )

#define FOREACH_OTHER_RESOURCE(resource) \
    for(resource = 64; resource < NO_OF_RESOURCES; resource++)

/*
**	The adjustment that needs to be made to the size and
**	offset fields for the unaligned access memory referencing
**	instructions. e.g. LDQU, STQU
*/
#define UNALIGNED_ACCESS_OFFSET_ADJUSTMENT -7
#define UNALIGNED_ACCESS_SIZE_ADJUSTMENT   7

#define UsingIPipe(instr) ((SCHIN_PIPELINES(instr) & PL_E0_E1) != 0)
#define UsingFPipe(instr) ((SCHIN_PIPELINES(instr) & PL_FA_FM) != 0)
#define IsPipelineAvailable(instr,AvPipes) \
       ((SCHIN_PIPELINES(instr) & AvPipes) != 0)

/*
**	-----------------------------------------------------------------
**	definitions used for propagating register latency information
**	between ** basic blocks.
**	-----------------------------------------------------------------
*/

/*
**  The number of basic block for which information about register
**  latencies at the beginning and ends must be kept.  (this should be
**  all "critical" basic blocks)
*/
/* #define CONFIG_NREGAVAIL 10000 */


typedef LATENCY REG_AVAILABLE [64];

    /* Comment on space requirements for BBL_REG_AVAILABLE: the 'start'
       field is probably not really useful as it is allways computed
       just before it is used.  The 'end' field is only useful for
       "critical" basic blocks. */

typedef struct {
   REG_AVAILABLE start;
   REG_AVAILABLE end;
   SLOT          first_free_slot;
   SLOT          first_free_slot_after;
} BBL_REG_AVAILABLE;

#define ABBL_REGAVAIL_START(bbl,reg)  bbl_reg_available [bbl].start [reg]
#define ABBL_REGAVAIL_END(bbl,reg)    bbl_reg_available [bbl].end   [reg]
#define ABBL_FIRST_FREE_SLOT(bbl)     bbl_reg_available [bbl].first_free_slot
#define ABBL_FIRST_FREE_SLOT_AFTER(bbl)  bbl_reg_available [bbl].first_free_slot_after
#define ABBL_REGAVAIL_RST(bbl)  bzero ((void*) bbl_reg_available, sizeof(bbl_reg_available))

#define ABBL_IS_SCHEDULED(bbl) (ABBL_FLAGS(bbl) & FB_SCHEDULED)

#define FAKE_INSTR_BEFORE_FIRST    SCHED_MAX_INSTRS-1
#define SCHED_LAST_POSSIBLE_INSTR  SCHED_MAX_INSTRS-2
